#! /usr/bin/env python3

import argparse
import graphviz
from pathlib import Path
import re

__version__ = "v1.0.0"
__author__ = "Axel VERDIER"
__organization__ = "INRAE LIPME BIOINFO team"
__contact__ = "axverdier@inrae.fr"

def is_code_file( filename: Path, valid_extensions: list[str] ) -> bool:
	return( filename.suffix in valid_extensions )

def get_all_code_files( directory: Path, recursive, valid_extensions: list[str] ) -> list[Path]:
	code_files = []
	for filename in directory.glob( "**/*" if recursive else "*" ):
		p = Path( filename )
		if is_code_file( p, valid_extensions ):
			code_files.append( p )
	return code_files

def get_includes( filename: Path ) -> list[str]:
	includes = []
	pattern = re.compile( "#include +\"(.*)\"" )
	with open( filename, 'r', encoding="ISO-8859-1" ) as f:
		for line in f:
			m = re.match( pattern, line )
			if m:
				include_name = Path( m.group(1) ).name # Use Path(...).name on the include to only get the file's name
				includes.append( include_name )
	return includes

def get_parameters() -> dict:
	parser = argparse.ArgumentParser(
						prog = "SourcesFilesGraphCreator",
						description = "A tool to create a graph of relation (includes) of sources files from a directory",
						epilog = '; '.join( [f"Version: {__version__}", f"Author: {__author__} ({__organization__})" ,f"Contact: {__contact__}" ] ),
						formatter_class = argparse.ArgumentDefaultsHelpFormatter
					)
	parser.add_argument( "sources_directory", help = "The directory to parse", )
	parser.add_argument( "-R", "--recursive", help = "Parse the directory recursively", action="store_true" )
	parser.add_argument( "-o", "--output_name", help = "The name, without extension, of the generated file", required = False, default = "graph" )
	parser.add_argument(
		"-e", "--extensions",
		help = "Add an extension to parser, you must include the dot (ex: '.hpp')",
		action="append", required = False , default = [ ".hpp", ".cpp", ".h", ".c"]
	)
	parser.add_argument(
		"-i", "--ignore",
		help = "Ignore a specific file",
		action = "append", required = False, default = []
	)

	args = parser.parse_args()

	return( {
		"sources_directory": Path( args.sources_directory ).absolute(),
		"recursive": args.recursive,
		"output_name": args.output_name,
		"extensions":  args.extensions,
		"ignored": args.ignore
	} )

if __name__ == "__main__":
	params = get_parameters()
	dot = graphviz.Digraph( comment = "Test comments" )

	# Get files list
	code_files = [ f for f in get_all_code_files( params["sources_directory"], params["recursive"], params["extensions"] ) if f.name not in params["ignored"] ]
	for f in code_files:
		dot.node( f.name )

	# Get links
	for f in code_files:
		for include in get_includes( f ):
			if include in params["ignored"]:
				continue
			same_name = Path(include).stem == f.stem
			# Note: look at https://www.graphviz.org/doc/info/attrs.html for possible attributes
			dot.edge(
				include,
				f.name,
				label = None,
				penwidth = "3" if same_name else "1",
				weight = "100" if same_name else "1"
			)

	# Create the graph
	dot.format = "svg"
	dot.render( params["output_name"], cleanup = True )
