// ------------------------------------------------------------------
// Copyright (C) 2004 INRA <eugene@ossau.toulouse.inra.fr>
//
// This program is open source; you can redistribute it and/or modify
// it under the terms of the Artistic License (see LICENSE file).
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//
// You should have received a copy of Artistic License along with
// this program; if not, please see http://www.opensource.org
//
// $Id$
// ------------------------------------------------------------------
// File:		 DNASeq.cc
// Contents: Definitions for a class representing DNA seq with ambiguous data
//	and 6 possible unambiguous versions (6 phases possible completions)
// ------------------------------------------------------------------


#ifndef  DNASEQ_H_INCLUDED
#define  DNASEQ_H_INCLUDED

/** \cond */ // Do not include standard depedencies in doxygen graph
#include <vector>
#include <string>
#include <map>
/** \endcond */

#include "logger.hpp"

const unsigned short MASKSEQ = 0x000F; // 0000 0000 0000 1111 

const unsigned short BitT = 3;
const unsigned short BitG = 2;
const unsigned short BitC = 1;
const unsigned short BitA = 0;


const unsigned short CodeT = (1 << BitT); // 000 000 000 1000
const unsigned short CodeG = (1 << BitG); // 000 000 000 0100
const unsigned short CodeC = (1 << BitC); // 000 000 000 0010
const unsigned short CodeA = (1 << BitA); // 000 000 000 0001

class  DNASeq {
	private:
		std::vector<unsigned short> Sequence;

		std::vector<std::string> vStart;	// vector of start codon
		std::vector<std::string> vStop;  // vector of start codon
		std::vector<std::string> vNonCanAcc; // vector of non canonical acceptor sites
		std::vector<std::string> vNonCanDon; // vector of non canonical donor sites
		std::map<std::string, std::string> mCodons; // mCodons["ATG"] = "M";

		void UpdateMarkov();
		void InitCodonTable();
		void InitNonCanSites();
		bool IsSite(const std::vector<std::string>&, size_t, int) const;

		Logger log;

		static char my_tolower( char c ) { return std::tolower( static_cast<unsigned char>(c) ); }

	public:
		DNASeq ();
		DNASeq (int size);
		DNASeq ( const std::string& filename);

		~DNASeq ();

		size_t SeqLen;
		std::string Name;

		double Markov0[4];

		void	Print(FILE *);
		void	PrintTr( FILE*, size_t, size_t, signed char );

		// Degeneracy returns the number of possible completely known sequence
		// represented by a degenerated subsequence
		unsigned char Degeneracy(size_t i, int sens, size_t len) const;
		double StartPenalty(int i, int sens);

		// bit vector values for spliced stop detection
		/*static const int isTf  = 1;
		static const int isTGf = 2;
		static const int isTAf = 4;
		static const int isTr  = 8;
		static const int isTGr = 16;
		static const int isTAr = 32;

		static const int isGf  = 1;
		static const int isGAf = 2;
		static const int isAf  = 4;
		static const int isARf = 8; // A or G
		static const int isGr  = 16;
		static const int isGAr = 32;
		static const int isAr  = 64;
		static const int isARr = 128;*/

		// bit vector values for spliced stop detection
		static const int startTf	= 1;
		static const int startAf	= 2;
		static const int startTGf = 4;
		static const int startTAf = 8;
		static const int startAGf = 16;
		static const int startTCf = 32;
		static const int startTr	= 64;
		static const int startAr	= 128;
		static const int startTGr = 256;
		static const int startTAr = 512;
		static const int startAGr = 1024;
		static const int startTCr = 2048;

		static const int stopAf  = 1;
		static const int stopGf  = 2;
		static const int stopGAf = 4;
		static const int stopAGf = 8;
		static const int stopAAf = 16;
		static const int stopGGf = 32;
		static const int stopCAf = 64;
		static const int stopAr  = 128;
		static const int stopGr  = 256;
		static const int stopGAr = 512;
		static const int stopAGr = 1024;
		static const int stopAAr = 2048;
		static const int stopGGr = 4096;
		static const int stopCAr = 8192;

		// see comments in .cc, used to detect possible spliced stops
		// returns a bit vector inside an int.
		int		 IsStartStop(size_t i) const;
		int		 IsStopStop(size_t i) const;

		// returns the fraction of completely known sequences that represent the
		// corresponding signal. Example: IsStop on TGN returns 0.25 (only TGA is a Stop)
		double IsAcc(size_t pos,int strand) const;
		double IsDon(size_t pos,int strand) const;
		double IsCanAcc(size_t pos,int strand) const;
		double IsCanDon(size_t pos,int strand) const;
		int		 IsNonCanAcc(size_t pos, int strand) const;
		int		 IsNonCanDon(size_t pos, int strand) const;

		double IsStop(size_t pos,int strand) const;
		double IsStart(size_t pos,int strand) const;
		double IsProStart(size_t i,int sens) const;
		double IsSpliceStart(size_t pos1, size_t pos2, int strand, int splicePos) const;
		bool IsAllowedStop(std::string triplet) const;
		bool IsAStop(std::string triplet) const;
		bool IsPattern(const char* site, size_t pos, int strand) const ;

		// Computes the markov probability of emission of the nuc. at position pos
		double Markov(size_t pos) const;
		// same for the reverse strand.
		double MarkovR(size_t pos) const;
		// returns the GC or AT% depending on the nuc. at position pos (GC or AT).
		double GC_AT(size_t pos) const;

		unsigned short Nuc2Code(char Ch) const;

		// copy from Pos to Len to To. mode is ambiguous or not. Len < 0 means reverse complement.
		void Transfer(int Pos, int Len, char *To, int mode);
		// compute the Frame (as in blast). Strand is '+' or anything else.
		int Pos2Frame(int pos, char strand) const;
		// remove ambiguity in the nuc. at pos
		unsigned short Unambit (int pos);

		char operator [] (size_t i) const; // reads DNA char at position i (forward)
		char operator () (size_t i) const; // reads DNA char at position i (complement)
		unsigned short operator () (size_t i, int mode) const ; // reads code at position i (mode: 0/forward, 1/complement, 2/revcomp)
		char AA(int i, int mode) const ; // translate the codon at position i, using mode as above.
		char nt(int i, int mode) const ; // dead code ?
};
#endif
