/***************************************************************************
    *   Copyright (C) 2007 by Noirot   *
    *   cnoirot@milieu   *
    *                                                                         *
    *   This program is free software; you can redistribute it and/or modify  *
    *   it under the terms of the GNU General Public License as published by  *
    *   the Free Software Foundation; either version 2 of the License, or     *
    *   (at your option) any later version.                                   *
    *                                                                         *
    *   This program is distributed in the hope that it will be useful,       *
    *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
    *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
    *   GNU General Public License for more details.                          *
    *                                                                         *
    *   You should have received a copy of the GNU General Public License     *
    *   along with this program; if not, write to the                         *
    *   Free Software Foundation, Inc.,                                       *
    *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "GeneFeature.hpp"
#include "GeneFeatureSet.hpp"

using namespace std;
// -----------------------
//  Default constructeur
// -----------------------
GeneFeature::GeneFeature()
{
	id_=".";
	seqid_=".";
	source_=".";
	type_=".";
	locus_=NULL;
	score_=0.0;
	phase_='.';
	attributes_=NULL;
	valid_= true ;
	length_=0;
}


//Constructor from a gff3 line
// prerequis : no \n at the end of the line.
GeneFeature::GeneFeature ( const char * const line, int lineno )
{
	id_=".";
	seqid_=".";
	source_=".";
	type_=".";
	locus_=NULL;
	score_=0.0;
	phase_='.';
	attributes_=NULL;
	valid_= true ;
	length_=0;

	const string line_str =to_string ( line );
	if ( line_str[0] != '#' ){
		// Chomp line ?
		size_t pos=line_str.find ( "\n" ) ;
		this->parseLine(
			pos == string::npos ? line_str : line_str.substr( 0, pos ),
			lineno
		);
	}
}

//Constructor from a gff3 line

GeneFeature::GeneFeature ( const GeneFeature & gene )
{
	id_=gene.id_;
	seqid_=gene.seqid_;
	source_=gene.source_;
	type_=gene.type_;
	if ( gene.locus_ != NULL )
	{
		locus_=new Locus ( *gene.locus_ );
	}
	score_=gene.score_;
	phase_=gene.phase_;
	if ( gene.attributes_ != NULL )
	{
		attributes_=new Attributes ( *gene.attributes_ );
	}
	valid_= gene.valid_ ;
	length_=gene.length_;
}
// ---------------------------------------
//  ParseLine : Parse a gff3 line with a token, check SO/SOFA code and Parent Attribut.
// ---------------------------------------

void GeneFeature::parseLine ( const string& line, int linenum )
{
	Logger log = Logger( "GeneFeature" );
	char delimiter = '\t';
	int i =0;
	int start = -1;
	int end = -1;
	char strand= '?';
	attributes_=NULL;
	locus_=NULL;
	length_=0;
	size_t token_start_pos = 0;
	size_t token_end_pos = line.find( delimiter );
	while ( token_start_pos != string::npos && i<9 ){ // Read the 9 tokens in line
		// Extract token
		const string token = line.substr(
			token_start_pos,
			token_end_pos == string::npos ? string::npos : token_end_pos - token_start_pos
		);

		//cout << "Token "<<i<< " valeur: "<<token<<endl;
		switch ( i )
		{
			case 0 :
			{
				//sequence id
				seqid_= token;
				break;
			}
			case 1 :
			{
				//source
				source_= token;
				break;
			}
			case 2 :
			{
				//type : SOFA/SO
				type_= token;
				if ( ( ! GeneFeatureSet::soTerms_->existsId ( type_ ) ) && ( !GeneFeatureSet::soTerms_->existsName ( type_ ) ) )
				{
					log.warning( type_  + " is not SO/SOFA referenced terms in line " + to_string(linenum) );
					valid_=false;
				}
				break;
			}
			case 3 :
			{
				start= stoi( token );
				break;
			}
			case 4 :
			{
				end= stoi( token );
				break;
			}
			case 5 :
			{
				try {
					// Note: previously atof was used.
					// If it fails, the score is set to 0.
					// and 0 makes sense in Eugene for GFF3 '.' score
					score_ = token == "." ? 0.0 : stod( token );
				} catch( const invalid_argument& e ) {
					string msg = "Couldn't convert score (" + token + ")";
					msg += " in line (" + to_string(linenum) + "):" + line;
					log.error( msg );
					exit( -1 );
				}
				break;
			}
			case 6 :
			{
				strand= token[0];
				break;
			}
			case 7 :
			{
				phase_= token[0];
				break;
			}
			case 8 :
			{
				//cout << "Attributes : " << token <<endl;
				attributes_ = new Attributes ( token );
				id_=attributes_->getId();
				if ( id_ == "" )
				{
					log.warning( "Feature has no ID in line " + to_string(linenum) );
					valid_=false;
				}
				break;
			}
		}

		// Move to next token, if any
		if ( token_end_pos == string::npos ) {
			token_start_pos = string::npos; // Will stop iteration
		} else {
			token_start_pos = token_end_pos + 1;
			token_end_pos = line.find( delimiter, token_start_pos );
		}
		i++;
	}
	locus_  = new Locus ( start, end, strand );
	length_ = end-start+1;
	if ( i < 9 )
	{
		log.warning( "incomplete GFF feature in line "  + to_string(linenum) );
		valid_ = false;
	}
}



// -----------------------
//  Destructeur
// -----------------------
GeneFeature::~GeneFeature ( )
{
	if ( locus_ != NULL )
	{
		delete locus_;
	}
	if ( attributes_ != NULL )
	{
		delete attributes_;
	}
}

// -----------------------
// Accessor
// -----------------------

// Sequence Id
//
void GeneFeature::setSeqId ( string seqid )
{
	seqid_ = seqid;
}
string GeneFeature::getSeqId ( ) const
{
	return seqid_;
}

// Id of feature , copy from attributes
//

string GeneFeature::getId ( ) const
{
	return id_;
}

// Parent attributes get from attributes class if exists
//
string GeneFeature::getParent() const
{
	string res="";
	if ( attributes_ != NULL )
	{
		res= attributes_->getParent();
	}
	return res;
}

// Type
//
string GeneFeature::getType() const
{
	return type_;
}

double GeneFeature::getScore() const
{
	return score_;
}
// Valid attribut.
//
void GeneFeature::setValid ( bool valid )
{
	valid_= valid;
}
bool GeneFeature::getValid ( ) const
{
	return valid_;
}
// ---------------------------------------
//  getString : Build a string gff3 format
// ---------------------------------------
string GeneFeature::getString ( ) const
{
	string geneFeature = seqid_+"\t"+source_+"\t"+type_+"\t";
	if ( locus_ != NULL )
	{
		geneFeature += to_string ( locus_->getStart() ) + "\t" + to_string ( locus_->getEnd() ) + "\t" + to_string ( score_ ) + "\t" + to_string ( locus_->getStrand() ) + "\t";
	}
	else
	{
		geneFeature += ".\t.\t" + to_string ( score_ ) + "\t.\t";
	}
	geneFeature += to_string ( phase_ ) +"\t";
	if ( attributes_ != NULL )
	{
		geneFeature +=attributes_->getString();
	}
	return geneFeature;
}

//Locus
Locus * GeneFeature::getLocus ( )  const
{
	return locus_;
}

char GeneFeature::getPhase ( ) const
{
	return phase_;
}

int GeneFeature::getLength ( ) const
{
	return ( length_ );
}

Attributes * GeneFeature::getAttributes ( ) const
{
	return ( attributes_ );
}

bool    GeneFeature::hasTarget ( )  const
{
	bool res=false;
	if ( attributes_ != NULL )
	{
		if ( attributes_->hasTarget() )
		{
			res=true;
		}
	}
	return res;
}
