// ------------------------------------------------------------------
// Copyright (C) 2004 INRA <eugene@ossau.toulouse.inra.fr>
//
// This program is open source; you can redistribute it and/or modify
// it under the terms of the Artistic License (see LICENSE file).
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of Artistic License along with
// this program; if not, please see http://www.opensource.org
//
// $Id$
// ------------------------------------------------------------------
// File:		 Hits.cc
// Contents: Definitions for a class representing alignements with genomic seq.
// ------------------------------------------------------------------

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/** \cond */ // Do not include standard depedencies in doxygen graph
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>
#include <algorithm>
#include <string>
/** \endcond */

#include "Hits.hpp"

using namespace std;

// ---------------------------------------------------------------------
//	Default constructor.
// ---------------------------------------------------------------------
Block :: Block (){
	Prev = nullptr;
	Next = nullptr;
}

// ---------------------------------------------------------------------
//	Constuct from Deb/Fin
// ---------------------------------------------------------------------
Block ::	Block(size_t start, size_t end, size_t lstart, size_t lend,int Ph,int Scr){
	Start  = start;
	End		 = end;
	LStart = lstart;
	LEnd	 = lend;
	Phase  = Ph;
	Score  = Scr;
	Prev = nullptr;
	Next = nullptr;
}

// ---------------------------------------------------------------------
//	Create a new Block and insert it after
// ---------------------------------------------------------------------
void Block ::  AddBlockAfter(size_t start, size_t end, size_t lstart, size_t lend,int Ph,int Scr)
{

		Block *ABlock = new Block(start,end,lstart,lend,Ph,Scr);
		this->Next		= ABlock;
		ABlock->Prev	= this;
}

// ---------------------------------------------------------------------
//	Insert newBlock after
// ---------------------------------------------------------------------
void Block::AddBlockAfter(Block* newBlock)
{
		this->Next		 = newBlock;
		newBlock->Prev = this;
}

// ---------------------------------------------------------------------
//	Compare the two blocks according to their genomic positions
// ---------------------------------------------------------------------
bool BlockPosCompare(const Block* a, const Block* b)
{
		return a->Start < b->Start;
}

// ---------------------------------------------------------------------
//	Destroy this Block and all blocks after him
// ---------------------------------------------------------------------
Block :: ~ Block	(){
	/* To avoid lot of recursive calls (and possible stack overflow) when deleting blocks:
	 * - Move forward on blocks: position on next block (which is now the current block) and get the next one
	 * - Make the current block independent to others (break chain)
	 * - Delete current block
	*/
	if( this->Next != nullptr ){
		Block* next_block_to_delete = this->Next;
		this->Next = nullptr;
		while( next_block_to_delete != nullptr ){
			Block* current_block_to_delete = next_block_to_delete;
			next_block_to_delete = current_block_to_delete->Next;
			current_block_to_delete->Next = nullptr; // Break chain to avoid recursive calls
			delete current_block_to_delete;
		}
	}
}

// ---------------------------------------------------------------------
//	Default constructor.
// ---------------------------------------------------------------------
Hits :: Hits	()
{
		Rejected = 0;
		Start		 = 0;
		End			 = 0;
		Evalue	 = 0.0;
		Level		 = 0;
		Support  = 0;
		Length	 = 0;
		Strand	 = 0;
		NGaps		 = 0;
		Name = "";
		Match = nullptr;
		Next = nullptr;
}

// ---------------------------------------------------------------------
//	Construct from a char*...
// ---------------------------------------------------------------------
Hits::Hits ( const std::string& name, int length, char strand, size_t deb, size_t fin, size_t ldeb, size_t lfin, int Ph, int Scr, double Prob, int level, int sup){
		Rejected = 0;
		Name = name;
		Strand = strand;
		Length = length;
		Start  = deb;
		End		 = fin;
		Evalue = Prob;
		Level  = level;
		Support= sup;
		Match  = new Block(deb,fin,ldeb,lfin,Ph,Scr);
		NGaps  = 0;
		Next	 = nullptr;
}

// ---------------------------------------------------------------------
//	Read a table of Hits from a file
// ---------------------------------------------------------------------
Hits* Hits::ReadFromFile(FILE* HitFile, int *NumHits, int level, int margin, size_t maxPos){
	Logger log = Logger( "Hits ReadFromFile" );
	char* HitId;
	char* PHitId;
	int	phase, Pphase, poids;
	double evalue/*, Pevalue*/;
	char	 A[512], B[512];
	char *HSP = NULL;
	Block *ThisBlock = NULL;
	Hits	*OneHit		 = NULL, *ThisHit = this, *AllHit = this;
	const int MaxHitLen = 15000;

	//Pevalue = -1.0;
	Pphase = 0;
	A[0]		= B[0] = 0;
	HitId		= A;
	PHitId	= B;

	if (ThisHit != NULL)
			for (int i=0; i<*NumHits-1; i++) ThisHit = ThisHit->Next;

	char line[2048];
	while( fgets( line, sizeof(line), HitFile) != NULL ){
		size_t deb, fin, HSPDeb, HSPFin;
		if( sscanf(
				line, "%lu %lu %d %lf %d %s %lu %lu",
				&deb, &fin, &poids, &evalue, &phase, HitId, &HSPDeb, &HSPFin
			) != 8
		){
			log.info( "Failed to parse HitLine " + string(line) );
			continue;
		}

		if (HSP) log.info( string(HSP) );
		if (phase < 0  &&  deb > fin){
			size_t tmp = deb;
			deb			= fin;
			fin			= tmp;
			tmp			= HSPDeb;
			HSPDeb	= HSPFin;
			HSPFin	= tmp;
		}

		if ((deb>maxPos) || (fin>maxPos)){
			log.info( "Hit does not map on sequence. Check " + string(HitId) );
			exit(1);
		}
		if ( ( fin > deb ? fin-deb : deb-fin ) > MaxHitLen){
			log.info( "Similarity of extreme length rejected. Check " + string(HitId) );
			continue;
		}
		if (
			(strcmp(HitId,PHitId) == 0) &&
			(phase*Pphase >= 0) &&
			(deb + margin > ThisBlock->End)  &&
			(
				( phase >= 0	&&	(HSPDeb + margin > ThisBlock->LEnd) )		 ||
				( phase < 0		&&	(HSPDeb - margin < ThisBlock->LEnd) )
			)
		){
					// si HitId et PHitId sont egaux, alors il y a un Hit en cours
					// de meme nom on verifie que c'est bien compatible en terme
					// de position (sur l'est et le genomique) et en e-value, en
					// prenant en compte le brin !
			ThisHit->NGaps++;
			ThisHit->End = fin-1;
			ThisBlock->AddBlockAfter(deb-1,fin-1,HSPDeb,HSPFin,phase,poids);
			ThisBlock = ThisBlock->Next;
		} else {
			(*NumHits)++;
			OneHit = new Hits(HitId, poids, (phase>0 ? '+' : '-'), deb-1, fin-1,
			HSPDeb, HSPFin, phase, poids, evalue, level, 0);
			ThisBlock = OneHit->Match;
			//CN Correct bug, assignment wasn't on data but on adress : PHitId = HitId
			strcpy (PHitId , HitId );
			//Pevalue		= evalue;
			Pphase = phase;
			if (AllHit == NULL){
				AllHit = OneHit;
				ThisHit = OneHit;
			} else {
				ThisHit->Next = OneHit;
				ThisHit				= OneHit;
			}
		}
	}
	return AllHit;
}

// ---------------------------------------------------------------------
//	Read a table of Hits from a file
// ---------------------------------------------------------------------
Hits* Hits::ReadFromGeneFeatureSet(GeneFeatureSet & HitSet , int *NumHits, int level, int /*margin*/, DNASeq *X ){
	Logger log = Logger( "Hits ReadFromGeneFeatureSet " + string(X->Name) );

	const int maxPos = X->SeqLen;
	Hits* ThisHit = this;
	Hits* AllHit = this;
	const int MaxHitLen = 15000;

	//	because this Hit mais be non empty at the method's start
	if (ThisHit != NULL && !(ThisHit->Name.empty()) ){
		for (int i=0; i<*NumHits-1; i++){
			ThisHit = ThisHit->Next;
		}
	}

	// Iterator to scan parent after parent
	map<string, vector<GeneFeature *> >::const_iterator itParent = HitSet.getIteratorParentToChildren();
	const int nbParents = HitSet.getNbParentFeature();

	for (int j=0 ; j < nbParents; j++, itParent++){
		vector<Block*> vBlocks;
		if (itParent->first == ""){ continue; }

		//Get SO code if feature correspond to the name or the synonym.
		string parentSo = HitSet.getGeneFeature(itParent->first)->getType();
		if ( parentSo.find("SO:") == string::npos ){
			parentSo = GeneFeatureSet::soTerms_->getIdFromName(parentSo);
		}

		string HitId;
		char hitStrand =	'?';
		double hitEvalue = -1;

		vector<GeneFeature *>::const_iterator it = itParent->second.cbegin();
		int nbGeneFeature = itParent->second.size();
		int i=0;
		// Create one block per child
		for ( i = 0 ; i < nbGeneFeature ; i++, it++ ){
			if ( ! (*it)->hasTarget() ){ continue; }
			int deb = (*it)->getLocus()->getStart();
			int fin = (*it)->getLocus()->getEnd();
			int poids  = (*it)->getAttributes()->getTarget()->getScoreHit();
			if ( poids <= 0){ poids = (*it)->getLength(); }
			char strand = (*it)->getLocus()->getStrand();
			HitId = (*it)->getAttributes()->getTarget()->getName();
			int HSPDeb = (*it)->getAttributes()->getTarget()->getLocus()->getStart();
			int HSPFin = (*it)->getAttributes()->getTarget()->getLocus()->getEnd();

			int phase = -1; // -1 for undefined
			if (parentSo == "SO:0000668"){ // EST_match
				phase  = (strand == '+' ? 0 : 1);
			} else { //BlastX
				phase  = (*it)->getAttributes()->getTarget()->getFrameHit();
				if (phase == 0){ //not specified
					phase  = X->Pos2Frame(deb,strand);
				} else {
					int computedFrame=X->Pos2Frame(deb,strand);
					if (phase != computedFrame){ //check frame computrd and read are the same.
						log.info(
							"Computed frame (" + to_string(computedFrame) + ") and input frame (" + to_string(phase) + ") are different. "+
							" Check :" +
							" " + to_string( HitId ) +
							" " + to_string( deb ) +
							" " + to_string( fin ) +
							" " + strand +
							" " + to_string( HSPDeb ) +
							" " + to_string( HSPFin )
						);
						continue;
					}
				}
			}
			if (phase < 0  &&  strand == '-' ){
				int tmp = HSPDeb;
				HSPDeb	= HSPFin;
				HSPFin	= tmp;
			}
			if ((deb>maxPos) || (fin>maxPos)){
				log.error( "Hit does not map on sequence. Check " + to_string(HitId) );
				exit(1);
			}
			if (abs(fin-deb) > MaxHitLen){
				log.error( "Similarity of extreme length rejected. Check " + to_string(HitId) );
				continue;
			}

			// first child of the hits: save the strand and the evalue
			if (i == 0){
				hitStrand = strand;
				hitEvalue = (*it)->getScore();
			}
			// Add a new block and save it in the vector
			vBlocks.push_back( new Block(deb-1,fin-1,HSPDeb,HSPFin,phase,poids )	);
		}
		if (vBlocks.size() > 0){
			// sort the block according to the genomic positions
			std::sort(vBlocks.begin(), vBlocks.end(), BlockPosCompare);
			// Get the first block
			Block* ThisBlock = vBlocks[0];
			// create a new hit and the first block of the hit
			Hits* OneHit = new Hits(
				HitId, ThisBlock->Score, hitStrand, ThisBlock->Start, ThisBlock->End, ThisBlock->LStart, ThisBlock->LEnd,
				ThisBlock->Phase, ThisBlock->Score, hitEvalue, level, 0
			);
			(*NumHits)++;
			ThisBlock = OneHit->Match;
			// Add other blocks to the hits and inc the number of gaps
			for (size_t k = 1; k< vBlocks.size(); k++){
				ThisBlock->AddBlockAfter(vBlocks[k]);
				ThisBlock = ThisBlock->Next;
				OneHit->NGaps++;
			}
			// Update the end position of the Hit = end position of the last block
			OneHit->End		= vBlocks[vBlocks.size()-1]->End;
			if (AllHit == NULL || AllHit->Name.empty() )	{
				AllHit	 = OneHit;
				ThisHit  = OneHit;
			} else {
				ThisHit->Next = OneHit;
				ThisHit				= OneHit;
			}
		}
	}
	return AllHit;
}


// ---------------------------------------------------------------------
//	Destroy this alignement
// ---------------------------------------------------------------------
Hits :: ~ Hits	(){
	delete this->Match;
	this->Match = nullptr;
	delete Next;
	Next = nullptr;
}
