// ------------------------------------------------------------------
// Copyright (C) 2004 INRA <eugene@ossau.toulouse.inra.fr>
//
// This program is open source; you can redistribute it and/or modify
// it under the terms of the Artistic License (see LICENSE file).
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//
// You should have received a copy of Artistic License along with
// this program; if not, please see http://www.opensource.org
//
// $Id$
// ------------------------------------------------------------------
// File:     MSensor.cc
// Contents: classes UseSensor, MasterSensor
// ------------------------------------------------------------------

#include <dlfcn.h>
#include <filesystem>

#include "MSensor.hpp"

using namespace std;
namespace fs = filesystem;

/*************************************************************
 **                       MasterSensor                      **
 *************************************************************/

extern Parameters PAR;

Logger MasterSensor::log = Logger( "MasterSensor" );

// ------------------------
// Sensors loading management
// ------------------------
map< string, SensorConstructor > MasterSensor::registered_sensors = {};
vector< void* > MasterSensor::plugins_sensors_handlers = {};

bool MasterSensor::declare_sensor( const string& sensor_name, const SensorConstructor& constructor, const bool lenient ){
  // Check this sensor is not already declared
  if( MasterSensor::registered_sensors.count( sensor_name ) ){
    if( lenient ){
      MasterSensor::log.warning( "Sensor " + sensor_name + " is already registered. Ignore this new one" );
      return false;
    } else {
      throw runtime_error( "New declaration of sensor " + sensor_name + " but it's already registered" );
    }
  }
  // Register this sensor
  MasterSensor::registered_sensors[sensor_name] = constructor;
  return true;
}

#ifdef ALLOW_PLUGINS
bool MasterSensor::load_plugin_sensor( const string& plugin_sensor_file ){
  void* handler = dlopen( plugin_sensor_file.c_str(), RTLD_NOW );
  char* error;
  if( (error = dlerror()) != NULL ){
    throw runtime_error(
      "Failed to load plugin sensor " + plugin_sensor_file +
      " with error: " + error
    );
  }
  MasterSensor::plugins_sensors_handlers.push_back( handler );
  MasterSensor::log.info( "Loaded plugin Sensor from " + plugin_sensor_file );
  return true;
}

void MasterSensor::load_all_plugin_sensors( const string& plugins_dir ){
  MasterSensor::log.info( "Plugins dir " + plugins_dir + ": loading sensors" );
  unsigned int counter = 0;
  for( const fs::directory_entry& entry: fs::recursive_directory_iterator( plugins_dir ) ){
    // Ignore non-file entries and extensions not ".so"
    if( !entry.is_regular_file() && entry.path().extension() != ".so" ){ continue; }
    // Load this file
    if( MasterSensor::load_plugin_sensor( fs::absolute( entry.path() ).string() ) ){
      counter++;
    }
  }
  MasterSensor::log.info( "Plugins dir " + plugins_dir + ": loaded " + to_string(counter) + " sensors" );
  return;
}
#else
bool MasterSensor::load_plugin_sensor( const string& ){
  MasterSensor::log.warning( "Not allowed to load plugins" );
  return false;
}

void MasterSensor::load_all_plugin_sensors( const string& ){
  MasterSensor::log.info( "Not loading plugins as it is not allowed in this eugene version" );
  return;
}
#endif

void MasterSensor::unregister_all_sensors() {
  MasterSensor::log.info( "Unregistering all sensors: start" );
  // Cleaning registered sensors
  MasterSensor::registered_sensors.clear();

  // Cleaning dynamic handlers
  for( void*& handler: MasterSensor::plugins_sensors_handlers ){
    dlclose( handler );
    handler = nullptr;
  }
  MasterSensor::plugins_sensors_handlers.clear();

  MasterSensor::log.info( "Unregistering all sensors: done" );
  return;
}

void MasterSensor::log_registered_sensors() {
	string msg = "Registered sensors:";
	for( const pair< const string&, SensorConstructor> kv: MasterSensor::registered_sensors ){
		msg += " " + kv.first;
	}
	MasterSensor::log.info( msg );
}

// ------------------------
//  Default destructor.
// ------------------------
MasterSensor::MasterSensor(){
}

MasterSensor::~MasterSensor (){
  MasterSensor::log.info( "Cleaning an instance: start" );
  for( Sensor* s: this->sensors ){
    MasterSensor::log.debug( "Deleting sensor on address " + to_string(s) );
    if( s != nullptr ){
      delete s;
      s = nullptr;
    } else {
      MasterSensor::log.warning( "Cleaning an instance: a stored Sensor pointer was already nullptr" );
    }
  }
  MasterSensor::log.info( "Cleaning an instance: all instances deleted, clearing list" );
  this->sensors.clear();
  MasterSensor::log.info( "Cleaning an instance: done" );
}



// ------------------------
//  Init Master.
// ------------------------
void MasterSensor :: InitMaster (DNASeq *X){
  MasterSensor::log.info( "Start initialization of a master with DNAseq: " + string(X->Name) );
  if( !this->is_initialized){
    // On r�cup�re les couples nom de sensor/priorit� du .par
    string key_name;
    int  val_prior;
    PAR.ResetIter();

    struct UseSensor{
      int priority;
      string name;
      UseSensor( int priority, const string& name ): priority(priority), name(name) {}
      static bool compare( const UseSensor& A, const UseSensor& B ){ return A.priority < B.priority; }
    };

    vector<UseSensor> msList = {};
    while( PAR.getUseSensor(key_name, val_prior) ){
      MasterSensor::log.info( "Got order to use sensor " + key_name + " with priority " + to_string(val_prior) );
      msList.emplace_back( val_prior, key_name );
    }

    // On les tri
    sort(msList.begin(), msList.end(), UseSensor::compare );
    
    // Update the list of sensors defined at the top level
    for( const UseSensor& use_sensor: msList ){
      this->sensors_name_ordered.push_back( use_sensor.name );
    }
    
    this->is_initialized = true;
  } 

	// create instance(s) of used sensors
	for( const string& sensor_name: this->sensors_name_ordered ){
		std::string use_name = "Sensor." + sensor_name + ".use";
		if (PAR.probeKey( use_name.c_str() )){
			for ( int j=0; j<PAR.getI( use_name.c_str() ); j++){
				MasterSensor::log.info( "Create and initialize an instance (id=" + to_string(j) + ") of the sensor " + sensor_name );
				sensors.push_back( MakeSensor( sensor_name.c_str(),j, X) );
				sensors.back()->Init(X);
			}
		}
	}
  MasterSensor::log.info( "Done initialization of a master with DNAseq: " + string(X->Name) );
}

// ------------------------
// Create an instance of a sensor and load the .so before if necessary
// ------------------------
Sensor* MasterSensor :: MakeSensor (std::string name, int n, DNASeq *X){
  if( !MasterSensor::registered_sensors.count( name ) ){
    string error_msg = "Tried to instanciate an unknown sensor: " + name;
    error_msg += "\nIf it's a plugin sensor, check:";
    error_msg += " the relative .so was provided";
    error_msg += ", the .so contains a valid sensor (inherit from Sensor) and a valid constructor";
    error_msg += ", the .so contains the declaration of the sensor to the master";
    throw runtime_error( error_msg );
  }
  return MasterSensor::registered_sensors.at( name )( n, X, this );
}

// ------------------------
//  Init. the sensors.
// ------------------------
void MasterSensor :: InitSensors (DNASeq *X){
  for( Sensor* sensor: this->sensors ){
    sensor->Init(X);
  }
}

// --------------------------
//  Get informations at pos.
// --------------------------
void MasterSensor :: GetInfoAt (DNASeq *X, int pos, DATA *d)
{
  int i;
  
  for(i=0; i < DATA::LastSigType;  i++)
    d->sig[i].Clear();
  
  for(i=0; i< DATA::LastContentsType; i++) d->contents[i] = 0.0;
  
  d->EstMatch = 0; // WARNING temporaire : EST -> on est dans intron

  for( Sensor* sensor: this->sensors ){
    sensor->GiveInfo( X, pos, d );
  }  
  
  for (i=0; i< DATA::LastSigType; i++)
    d->sig[i].SetToDefault();
}

// --------------------------
//  Print informations at pos.
// --------------------------
void MasterSensor :: PrintDataAt (DNASeq *X, int pos, DATA *d, FILE *OUT)
{
  int i,j;
  fprintf(OUT, "%6d %c", 1+pos, (*X)[pos]);

  for(i=0; i<DATA::LastContentsType; i++)
    fprintf(OUT, " %5.3g",d->contents[i]);

  for(i=0; i< Signal::LastEdge;  i++) {
    fprintf(OUT, " ||");
    for (j=0; j< DATA::LastSigType; j++)
      fprintf(OUT, " %5.3g",d->sig[j].weight[i]);
  }
  fprintf(OUT, "\n");
}

// --------------------------------------------
//  Get special info at special pos.
//  Retourne 1 si les sensors sont porteurs
//  d'infos de type "type" 0 sinon.
// --------------------------------------------
int MasterSensor :: GetInfoSpAt (unsigned char type,
				 DNASeq *X, int pos, DATA *d)
{
  int i;
  int info = 0;  // Aucune info
  
  for(i=0; i< DATA::LastSigType;  i++)
    d->sig[i].Clear();
  
  for(i=0; i<DATA::LastContentsType; i++) d->contents[i] = 0.0;
  
  d->EstMatch = 0; // WARNING temporaire : EST -> on est dans intron

  for( Sensor* sensor: this->sensors ){
    if ( sensor->type & type) {
      sensor->GiveInfo(X, pos, d);
      info = 1;
    }
  }

  for (i=0; i< DATA::LastSigType; i++)
    d->sig[i].SetToDefault();

  return info;
}

// --------------------------
//  Post analyse the sensors.
// --------------------------
void MasterSensor :: PostAnalyse (Prediction *pred, FILE *MISC_INFO){
  for( Sensor* sensor: this->sensors ){
    sensor->PostAnalyse( pred, MISC_INFO );
  }
  return;
}

