// ------------------------------------------------------------------
// Copyright (C) 2004 INRA <eugene@ossau.toulouse.inra.fr>
//
// This program is open source; you can redistribute it and/or modify
// it under the terms of the Artistic License (see LICENSE file).
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//
// You should have received a copy of Artistic License along with
// this program; if not, please see http://www.opensource.org
//
// $Id$
// ------------------------------------------------------------------
// File:     MSensor.hpp
// Contents: classes UseSensor, MasterSensor
// ------------------------------------------------------------------

#ifndef  MSENSOR_H_INCLUDED
#define  MSENSOR_H_INCLUDED

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/** \cond */ // Do not include standard depedencies in doxygen graph
#include <cstdio>
#include <vector>
#include <algorithm>
#include <string>
#include <map>
/** \endcond */

#include "Sensor.hpp"

typedef Sensor*(*SensorConstructor)(int,DNASeq*,MasterSensor*); // Define SensorConstructor type for function: Sensor fct( int, DNASeq*, MasterSensor* ){}


/*************************************************************
 **                      MasterSensor                       **
 *************************************************************/
/*! Manage sensors */
class MasterSensor {
	private:
		static Logger log;
		static std::map< std::string, SensorConstructor > registered_sensors; /*!< All known sensors (intern and plugins) and their constructor function */
		static std::vector< void* > plugins_sensors_handlers;  /*!< All plugins sensor handlers (dynamic loaded) */

		std::vector<std::string> sensors_name_ordered; /*!< Sensors' name (sorted) to use */
		std::vector<Sensor*> sensors; /*!< Sensors' instances of this MasterSensor */
		bool is_initialized = false;

	public:
		MasterSensor();
		~MasterSensor();

		/* ========== Sensors (un)loading management ========== */
		/*! \brief Method to use by a sensor to declare its existence
		* \param sensor_name  The name of the sensor
		* \param constructor  The constructor function (building and returning an instance of the sensor)
		* \param lenient      Control how the function manage failure to declare the sensor: true => return false, false => raise an error
		*/
		static bool declare_sensor( const std::string& sensor_name, const SensorConstructor& constructor, const bool lenient = true );

		/*! Method to load and declare (not instanciate) 1 plugin sensor from it's .so file */
		static bool load_plugin_sensor( const std::string& plugin_sensor_file );

		/*! Method to load and declare (not instanciate) all plugin sensors (.so files) present in a directory */
		static void load_all_plugin_sensors( const std::string& plugins_dir );

		/*! Method to clean declared sensor, **should be call at the end of sensors usage and only once** */
		static void unregister_all_sensors();

		/*! Method log (INFO level) registered sensors, do not modify anything */
		static void log_registered_sensors();

		/* ========== Sensors' instances management ========== */
		void InitMaster   (DNASeq *X);
		void InitSensors  (DNASeq *X);
		void GetInfoAt    (DNASeq *X, int pos, DATA *d);
		void PrintDataAt  (DNASeq *X, int pos, DATA *d,   FILE *OUT=stdout);
		int  GetInfoSpAt  (unsigned char type, DNASeq *X, int pos, DATA *d);
		void PostAnalyse  (Prediction *pred,   FILE *MISC_INFO);
		Sensor* MakeSensor (std::string name, int n, DNASeq *X);  // create an instance of a sensor
};



#endif
