// ------------------------------------------------------------------
// Copyright (C) 2004 INRA <eugene@ossau.toulouse.inra.fr>
//
// This program is open source; you can redistribute it and/or modify
// it under the terms of the Artistic License (see LICENSE file).
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
//
// You should have received a copy of Artistic License along with
// this program; if not, please see http://www.opensource.org
//
// $Id: Prediction.h,v 1.49 2012-11-07 09:44:47 sallet Exp $
// ------------------------------------------------------------------
// File:     Prediction.h
// Contents: class Prediction
// ------------------------------------------------------------------

#ifndef  PREDICTION_H_INCLUDED
#define  PREDICTION_H_INCLUDED

/** \cond */ // Do not include standard depedencies in doxygen graph
#include <cstdio>
#include <vector>
#include <algorithm>
#include <fstream>
#include <iostream>
/** \endcond */

#include "AltEst.hpp"
#include "Prediction_cte.h"
#include "Gff3Line.hpp"
#include "MSensor.hpp"
#include "State.hpp"

/*************************************************************
 **                       Feature Object                    **
 *************************************************************/
class Feature {
	private:
	public:
		int    number;
		int    start;
		int    end;
		char   strand;
		int    phase;
		int    frame;
		State* featureState;
		int    cds_start; // if feature is a cds, cds start position (can be different from start if there is a FS)
		int    cds_end;   // if feature is a cds, cds end position (can be different from end if there is a FS)
  
		Feature  ();
		Feature  (signed char, int, int);
		~Feature ();
		bool      Overlap (const Feature&);
		bool      IsCodingExon();
		bool      IsIntron();
		bool      IsUTR();
		bool      IsUTRIntron();
		bool      IsIntergenic();
		bool      IsTranscribedAndUnspliced();
		bool      IsNpcRna();
		bool      IsBicoding();
		void      ComputePhase(int);
		void      SetState(char);
};


/*************************************************************
 **                         Gene Object                     **
 *************************************************************/
class Gene {
	private:
		int  complete;  //+1:init|sngl +2:term|sngl 3:ALL
		int  exNumber;
		int  inNumber, inLength;
		int  utrLength, mrnaLength, geneLength;
		char strand;
		bool isNpcRna;   // true if it is a non protein coding rna
		int  operonNb; // operon number
		//std::vector<int> vFrameShift; TODO: save the positions of the frameshift in a array
		int hasFrameShift;
		void clear    ();
	public:
		int cdsStart, cdsEnd, trStart, trEnd;
		int tuStart, tuEnd; // transcript unit (required for gff3 + alternatives variants)
		int exLength;
		int geneNumber;
		bool isvariant; // true = is a splice variant
		int hasvariant; // has n splice variant (or is number n splice variant if is variant is true)
		std::vector <Feature*> vFea;

		Gene  ();
		Gene (std::string line, DNASeq& seq);
		~Gene ();
		void Update(int);

		int isDifferent (const Gene& o, int threshold);
		bool operator== (const Gene& o);
		bool HasSameExons(const Gene& o);
		bool Overlap (const Gene& o);
		float GetOverlapWith ( const Gene& g) ;
		inline int nbFea() {return vFea.size();};
		int  GetExonNumber();
		char GetStrand();
		int  GetOperonNb();
		void SetOperonNb(int);
		void AddFeature(signed char state, int start, int end);
		void PrintInfo (FILE*, int, char*);
		void Print();
		std::string GetVariantCode(void) const;
		bool IsNpcRna();
		void ManageFrameShift(int);
		void ConvertToNpcRNA(int);
		int get_complete() const { return this->complete; }
		int get_mrnaLength() const { return this->mrnaLength; }
		int get_geneLength() const { return this->geneLength; }
		char get_strand() const { return this->strand; }
		int get_exNumber() const { return this->exNumber; }
		int get_hasFrameShift() const { return this->hasFrameShift; }
};

/*************************************************************
 **                         Operon Object                     **
 *************************************************************/
class Operon {
	private:
		int  start;
		int  end;
		char strand;
		int  number;
	public:
		std::vector <Gene*> vGenes;
		Operon();
		Operon(int nb);
		~Operon();
		int  GetStart();
		int  GetEnd();
		int  GetNumber();
		char GetStrand();
		void SetNumber(int );
		void AddGene(Gene* gene);
		void Print();
};


/*************************************************************
 **                        Prediction                       **
 ************************************************************/
class Prediction {
	private:
		DNASeq       *X;
		std::vector<unsigned char> ESTMatch;

		inline unsigned char GetESTMatch(size_t i) {return (ESTMatch.size() > i ? ESTMatch[i] : 0);};
		void  clear     ();
		void  PrintGff3        (std::ofstream&, char*, char, MasterSensor* );
		void  PrintEgnL       (FILE*, char*, MasterSensor*, int a=0 );
		void  PrintEgnS       (FILE*);
		void  PrintEgnD       (FILE*, MasterSensor*);
		void  PrintHtml       (FILE*, char*);

		// Verif coherence EST: calcul le nombre de nuc. coherents et
		// incoherents avec les match est
		// debut/fin/etat: debut et fin de la seq. dont l'etat est etat
		// cons/incons: retour des valeurs
		void  CheckConsistency(int debut, int fin, int etat,
		int* cons, int* incons, MasterSensor* ms);

		// Check and Trim UTR according to EST evidence
		void ESTScan();
		bool UTRCheckAndTrim(int* debut, int* fin, int etat);
		void UpdateAndDelete();

		// PrintHtml : -ph print the start of the HTML output
		void  StartHTML       (char*, FILE*);

		// PrintHtml : -ph print the end of the HTML output
		void  EndHTML         (FILE*);

		Gff3Line* fillGff3Line(int type_sofa, int start, int end,
								char strand, int phasegff);
		void setGff3Attributes(Gff3Line* line, State* featState,
						int type_sofa, std::string fea_name,
						int j, std::string code_variant, std::string gene_id, bool coding);
		bool previousExonMustBeUpdated(Gff3Line* line, int start);


		// -------------------------------------------------------------------------
		// Initialize the active tracks vector.
		// At the moment, just copy the values including in the array proActiveTracks
		// or eukActiveTracks in Prediction_cte.h
		// TODO Load a file and remove these arrays
		// -------------------------------------------------------------------------
		/** Get the active tracks based on current mode (from Parameters) and the strand
		 * @return	A 2-tuple, the first element is the active tracks and the second the indication of which track is active
		 */
		static std::tuple< std::vector<short int>, std::vector<bool> > get_active_tracks( int strand );

	public:
		std::vector <Gene*>   vGene;
		std::vector <Operon*> vOperon;
		int    nbGene;        // Number of predicted genes (including gene non coding for protein)
		char   seqName[FILENAME_MAX+1];
		double optimalPath;

		Prediction  ();
		Prediction  (int From, int To, std::vector <int> vPos,
         std::vector <signed char> vState);
		Prediction  (const std::string & desc, DNASeq*);
		Prediction  (char [FILENAME_MAX+1], DNASeq *);
		~Prediction ();

		void LoadGene(std::vector<GeneFeature*>&, std::vector<int>&, std::vector<signed char>&, DNASeq*);
		void Init (int From, int To, const std::vector <int>& , const std::vector <signed char>&);
		bool IsOriginal(Prediction* optPred, std::vector <Prediction*>& altPreds, int seuil) const;
		void  TrimAndUpdate (DNASeq*);
		void  SanityCheck();
		void  DeleteOutOfRange(int s,int e, char strand);
		void  Print         (DNASeq*, MasterSensor*, FILE *OTP_OUT=NULL, char append = 0, char variant = 0);
		void  PrintGeneInfo (FILE*);
		void  PlotPred      ();
		State* GetStateAtPos(int) const;
		Gene *FindGene(int start, int end, char strand = 0) const;
		bool compatibleWithAltEst( const OneAltEst& altEst ) const;
		Gene* GetUncompatibleGene ( const OneAltEst& altEst, int margin=0) const;
		void Print();
		std::vector<int> Eval(Prediction* ref, int offset, bool onlyCodingGene);
		std::vector<int> EvalGene(Prediction* ref, int start, int end, bool onlyCodingGene);
		std::vector<int> EvalExon(Prediction* ref, int start, int end);
		std::vector<Feature*> GetExons(int start, int end) const;
		std::vector<Gene*>    GetGenes(int start, int end, bool onlyCodingGene) const;
		void  ComputeOperons();
		Operon* GetOperon(int nb) const;
		int GetExonNumber() const;
		int GetExonLength() const ;
		void AppendPred(const Prediction*);

		// Need by Sensor Tester
		const char* IsStart       (int) const;
		const char* IsStop        (int) const;
		const char* IsDon         (int) const;
		const char* IsAcc         (int) const;
		bool  IsState       (DATA::SigType sig_type, int pos, char strand) const;

		static Prediction* Predict(DNASeq* TheSeq, int From, int To, MasterSensor* MSensor, int strand=0);
		static Prediction* Predict (DNASeq* TheSeq, MasterSensor* MSensor); // Same with whole sequence by default
		static Prediction* AltPredict (DNASeq* TheSeq, int From, int To, MasterSensor* MSensor, AltEst *AltEstDB, Prediction *optpred, int idx);
};


// Sensor Contents used by each track
const enum DATA::ContentsType SensorContents[NbTracks] = {
	DATA::ExonF1, DATA::ExonF2, DATA::ExonF3,
	DATA::ExonR1, DATA::ExonR2, DATA::ExonR3,
	DATA::ExonF1, DATA::ExonF2, DATA::ExonF3,
	DATA::ExonR1, DATA::ExonR2, DATA::ExonR3,
	DATA::ExonF1, DATA::ExonF2, DATA::ExonF3,
	DATA::ExonR1, DATA::ExonR2, DATA::ExonR3,
	DATA::ExonF1, DATA::ExonF2, DATA::ExonF3,
	DATA::ExonR1, DATA::ExonR2, DATA::ExonR3,
	DATA::IntronF, DATA::IntronF, DATA::IntronF,
	DATA::IntronR, DATA::IntronR, DATA::IntronR,
	DATA::IntronF, DATA::IntronF, DATA::IntronF,
	DATA::IntronF, DATA::IntronF, DATA::IntronF,
	DATA::IntronR, DATA::IntronR, DATA::IntronR,
	DATA::IntronR, DATA::IntronR, DATA::IntronR, DATA::IntronR,
	DATA::InterG,
	DATA::UTR5F, DATA::UTR3F,
	DATA::UTR5R, DATA::UTR3R,
	DATA::IntronUTRF, DATA::IntronUTRR,
	DATA::IntronUTRF, DATA::IntronUTRR,
	DATA::RNAF, DATA::RNAR,
	DATA::ExonF1, DATA::ExonF1, DATA::ExonF2, // Bicoding: Ces valeurs ne doivent jamais etre appelées
	DATA::ExonR1, DATA::ExonR1, DATA::ExonR2, // Bicoding: Ces valeurs ne doivent jamais etre appelées
	DATA::ExonF1, DATA::ExonF1, DATA::ExonF1, // Bicoding: Ces valeurs ne doivent jamais etre appelées
	DATA::ExonF2, DATA::ExonF2, DATA::ExonF2, // Bicoding: Ces valeurs ne doivent jamais etre appelées
	DATA::ExonF3, DATA::ExonF3, DATA::ExonF3, // Bicoding: Ces valeurs ne doivent jamais etre appelées
	DATA::UIRF, DATA::UIRR
};


#endif
