// ------------------------------------------------------------------
// Copyright (C) 2004 INRA <eugene@ossau.toulouse.inra.fr>
//
// This program is open source; you can redistribute it and/or modify
// it under the terms of the Artistic License (see LICENSE file).
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of Artistic License along with
// this program; if not, please see http://www.opensource.org
//
// $Id: EuGene.cc,v 1.109 2015-07-30 14:19:36 sallet Exp $
// ------------------------------------------------------------------
// File:     EuGene.cc
// Contents: This program finds exons/introns and intergenic regions
//           (including UTRs)
// ------------------------------------------------------------------


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/** \cond */ // Do not include standard depedencies in doxygen graph
#include <algorithm> // min, max
#include <string>
/** \endcond */

using namespace std;

#include "Param.hpp"
#include "Prediction.hpp"

#ifndef FILENAME_MAX
#define FILENAME_MAX        1024
#endif

// ------------------ Globals --------------------
Parameters       PAR;

void core_run ( const string& sequence, const bool graph ){
    Prediction *pred;
    string grname;

    PAR.set("fstname", sequence.c_str() );

    // --------------------------------------------------------------------
    // Lecture de la sequence
    // --------------------------------------------------------------------
    DNASeq seq = DNASeq( string(PAR.getC("fstname")) );
    const int seq_len = seq.SeqLen;

    Logger log = Logger( "core run on " + string(seq.Name) );
    log.info( "START" );

    // --------------------------------------------------------------------
    // Calcul des positions de prédiction
    // --------------------------------------------------------------------
    const int fromPos = PAR.getI("EuGene.from",0,true);
    const int toPos   = ( PAR.getI("EuGene.to",0,true) ) ? min( seq_len - 1, PAR.getI("EuGene.to",0,true) ) : seq_len - 1;

    // --------------------------------------------------------------------
    // Prefix output file name
    // --------------------------------------------------------------------
    string prefixName;
    prefixName =PAR.getC("Output.Prefix");
    prefixName += BaseName( PAR.getC("fstname") );

    const size_t prefixName_dot_pos = prefixName.rfind('.');
    if( prefixName_dot_pos != string::npos ){
        const string extension = prefixName.substr( prefixName_dot_pos );
        const vector<string> known_extensions =  { ".fasta", ".fsa", ".tfa", ".txt" };

        // Remove extension
        if( find( known_extensions.cbegin(), known_extensions.cend(), extension ) != known_extensions.cend() ){
            prefixName.erase( prefixName_dot_pos );
        } 
    }
    PAR.set("prefixName", prefixName.c_str());

    // --------------------------------------------------------------------
    // Preparation sortie graphique + Scores
    // --------------------------------------------------------------------
    if (graph)
    {
        int gto       = PAR.getI("Output.gto");
        int gfrom     = PAR.getI("Output.gfrom");
        int glen      = PAR.getI("Output.glen");

        // Construction du nom de sortie (*.png)
        grname = prefixName;

        if ((gfrom <= 0)|| (gfrom >= seq_len))
            gfrom = 1;
        if ((gto <= 0)  || (gto <= gfrom) || (gto > seq_len))
            gto = seq_len;

        if ((PAR.getI("Output.gfrom")!=-1) || PAR.getI("Output.gto")!=-1)
        {
            /*
            sprintf(grname+strlen(grname), ".%d", gfrom);
            sprintf(grname+strlen(grname), "-%d", gto);
            */
           grname += "." + to_string(gfrom);
           grname += "-" + to_string(gto);
        }

        gfrom--;
        gto--;

        if (glen < 0)
            glen = ((gto-gfrom+1 <= 6000) ? gto-gfrom+1 : 6000);

        InitPNG(PAR.getI("Output.resx"),   PAR.getI("Output.resy"),
                PAR.getI("Output.offset"), gfrom, gto,
                PAR.getI("Output.golap"), glen, grname.c_str());
    }

    // --------------------------------------------------------------------
    // Init MasterSensor
    // --------------------------------------------------------------------
    MasterSensor MS = MasterSensor();
    MS.InitMaster(&seq);

    // --------------------------------------------------------------------
    // Predict: 1st main prediction
    // --------------------------------------------------------------------
    if (PAR.count( strdup("AltEst.reference") ) > 0){
        log.info( "Variant prediction..." );
        const string reffile = PAR.getC("AltEst.reference");
        pred = new Prediction(reffile.c_str(), &seq);
    } else {
        const string mode = PAR.getC ("EuGene.mode");

        if ( mode == "Prokaryote2" || mode == "Eukaryote2" ){
            // predict on the forward strand
            log.info( "Running prediction on the forward strand" );
            Prediction* pred2 = Prediction::Predict(&seq, fromPos, toPos, &MS,  1);
            // predict on the reverse strand
            log.info( "Running prediction on the reverse strand" );
            pred = Prediction::Predict(&seq, fromPos, toPos, &MS, -1);
            pred->AppendPred(pred2);
        } else {
            log.info( "Running prediction" );
            pred = Prediction::Predict(&seq, fromPos, toPos, &MS);
            log.info( "Optimal path length =" + log.format( - pred->optimalPath ) );
        }
    }

    // --------------------------------------------------------------------
    // Textual and graphical output
    // --------------------------------------------------------------------
    if (graph){ pred->PlotPred(); }

    // Print the prediction in the classical mode or in the AltEst mode (without ref prediction preloading)
    if ( ! PAR.getI("AltEst.use") || (PAR.getI("AltEst.use") && PAR.count( strdup("AltEst.reference") ) <= 0 )){
        log.info( "Writing prediction" );
        pred->Print(&seq, &MS);
    }

    if (graph){
        log.info( "Dumping images (\"" + string(grname) + ".---.png\"): start" );
        ClosePNG();
        log.info( "Dumping images (\"" + string(grname) + ".---.png\"): done" );
    }
    
    // --------------------------------------------------------------------
    // Load Alternative EST data (if any)
    // --------------------------------------------------------------------
    if (PAR.getI("AltEst.use")){
        log.info( "Loading alternative EST data" );
        AltEst *AltEstDB = new AltEst(&seq);
        std::vector <Prediction*> vPred;
        const int ExonBorderMatchThreshold = PAR.getI("AltEst.ExonBorderMatchThreshold");
        const int RepredictMargin          = PAR.getI("AltEst.RepredictMargin");
        const int ExonBorderMatchThreshold2 = PAR.getI("AltEst.IncompatibilityExonBorderMatchThreshold");
        
        Prediction*               AltPred;
        Gene*                     uncompatibleGene;
        int localFrom,localTo, altLocalFrom,altLocalTo ;
        int NumberLaunchedPred = 0;

        log.info( "Running prediction on alternative EST data" );
        for (int altidx = 0; altidx < AltEstDB->totalAltEstNumber; altidx++){
            if (AltEstDB->voae_AltEst[altidx].IsToRemove()) continue;
            
            // Look for an overlapping with an inconsistent gene
            uncompatibleGene =  pred->GetUncompatibleGene( AltEstDB->voae_AltEst[altidx], ExonBorderMatchThreshold2 );
            // skip if not reference gene found
            if (uncompatibleGene == NULL){ continue; }
            NumberLaunchedPred++;
                
            // largest positions between altEst hit ones and gene ones
            altLocalFrom = min(uncompatibleGene->trStart, AltEstDB->voae_AltEst[altidx].GetStart());
            altLocalTo   = max(uncompatibleGene->trEnd,   AltEstDB->voae_AltEst[altidx].GetEnd());
            // positions with a margin
            localFrom = max(fromPos, altLocalFrom - RepredictMargin);
            localTo   = min(toPos,   altLocalTo + RepredictMargin);
                
            AltPred = Prediction::AltPredict(&seq,localFrom,localTo,&MS,AltEstDB,pred,altidx);

            if (AltPred){
                if ( (AltPred->vGene[0]->cdsStart == -1) || (AltPred->vGene[0]->cdsEnd == -1))
                {
                    delete AltPred;
                    continue;
                }
                // Delete the gene of the alt prediction which doesn't overlap the EST
                AltPred->DeleteOutOfRange(AltEstDB->voae_AltEst[altidx].GetStart(),AltEstDB->voae_AltEst[altidx].GetEnd(), AltEstDB->voae_AltEst[altidx].GetStrand());
                
                // If a gene overlapping the EST was found, if more than 80% of the variant overlaps its reference gene
                // And if the prediction is original, keep it
                if ( (AltPred->nbGene > 0) && (AltPred->vGene[0]->GetOverlapWith(*uncompatibleGene) >= MIN_PERCENTAGE_OVERLAP_ALTEST) && (AltPred->IsOriginal(pred,vPred,ExonBorderMatchThreshold)) ){
                    log.info( "Optimal path length = " + log.format( - AltPred->optimalPath, 4 ) );
                    uncompatibleGene->hasvariant++;
                    AltPred->vGene[0]->isvariant = true;
                    AltPred->vGene[0]->hasvariant = uncompatibleGene->hasvariant;
                    AltPred->vGene[0]->geneNumber = uncompatibleGene->geneNumber;
                    uncompatibleGene->tuStart = ( uncompatibleGene->tuStart ) ? min(uncompatibleGene->tuStart,AltPred->vGene[0]->trStart)
                                            : min(uncompatibleGene->trStart,AltPred->vGene[0]->trStart);
                    uncompatibleGene->tuEnd   = ( uncompatibleGene->tuEnd )   ? max(uncompatibleGene->tuEnd,AltPred->vGene[0]->trEnd)
                                            : max(uncompatibleGene->trEnd,AltPred->vGene[0]->trEnd);
                    vPred.push_back(AltPred);
                } else{
                     delete AltPred;
                }
            }
        }
        
        // print the prediction and variants
        log.info( "Writing predictions on alternative EST data" );
        pred->Print(&seq, &MS, NULL, 0, 1);
        for (size_t idx = 0; idx < vPred.size(); idx++){
            vPred[idx]->Print(&seq, &MS,NULL,1, 1);
        }
        for (size_t idx = 0; idx < vPred.size(); idx++){
            delete (vPred[idx]);
        }
    }

    // Free used memory
    delete pred;

    log.info( "FINISHED" );

    return;
}

// -------------------------------------------------------------------------
//            MAIN
// -------------------------------------------------------------------------
int main  (int argc, char * argv []){
    Logger log = Logger();
    const string logs_separator = "---------------------------------------------------------------------";

	log.debug( "DEBUG_MODE is on" );

    log.info( logs_separator );

    // Lecture de la ligne d'arg et du fichier .par
    PAR.initParam(argc, argv);

    log.info( logs_separator );

    // Registered sensors (load plugings)
    MasterSensor::load_all_plugin_sensors( (std::string)PAR.getC("eugene_dir")+"/"+PLUGINS_DIR+"/" );
	MasterSensor::log_registered_sensors();

	log.info( logs_separator );

    if (PAR.getI("ParaOptimization.Use")){
		log.warning( "Parameters Optimization mode has been remove" );
    } else {
        // Objectif : limiter les appels à la MAP
        bool output_graph = ( PAR.getI("Output.graph") > 0 );

        for (int sequence = optind; sequence < argc ; sequence++){
            core_run( argv[sequence], output_graph );
        } // fin de traitement de chaque séquence....

        log.info( logs_separator );
    }

    // Unload all sensors (intern and plugins)
    MasterSensor::unregister_all_sensors();
    return 0;
}
