#! /usr/bin/env python3

import csv
from datetime import datetime
import filecmp
import glob
import multiprocessing as mp
import os
import pathlib
import re
import subprocess
import sys
import time
import typing

NB_EUGENE_PARALLEL_RUNS: int = 5

# Test bulk infos
script_dir = pathlib.Path( __file__ ).parent.absolute()
integrations_tests_list_file = ( script_dir / "integration_tests.csv" ).resolve()
isodate = datetime.now().strftime( "%Y-%m-%d-%H-%M" )
egn_version = "v5.0.0"
egn_bin = ( script_dir / "../../bin/eugene" ).resolve()
egn_dir = ( script_dir / "../../" ).resolve() # Where are stored the configs of eugene
test_dataset_srv = "147.100.179.77"
test_dataset_dir = "/mnt/beegfs/work/eugene/tests/integration"

# Utils

def info( content: str ) -> None:
	print( "[INFO] " + content )
	return

def ok( content: str ) -> None:
	print( "[OK] " + content )
	return

def failed( content: str ) -> None:
	print( "[FAILED] " + content )
	return

def prepare_run_data( direction: str, data_dir: str, run_dir: str ) -> str:
	run_data_dir: str = f"{run_dir}/data{direction}"
	os.makedirs( run_data_dir, exist_ok=False )

	# Symlink all data
	for filename in os.listdir( data_dir ):
		if os.path.isfile( f"{data_dir}/{filename}" ):
			os.symlink( f"{data_dir}/{filename}", f"{run_data_dir}/{filename}" )

	# Symlink blast files for eugene
	for f in glob.glob( f"{run_data_dir}/*.blast*.{direction}.gff3"):
		suffix_size: int = 6 + len( direction )
		dest = f[:-suffix_size] + ".gff3"
		os.symlink( f, dest )

	return run_data_dir

def cmp_gff3s( gff3file_1: str, gff3file_2: str ) -> bool:
	same_files = True
	file_pairs = [ ( gff3file_1, gff3file_2 ), ( gff3file_2, gff3file_1 ) ]
	for f1, f2 in file_pairs:
		# f1 has more data than f2
		result = subprocess.run(
			[
				"bedtools", "subtract",
				"-a",  f1, "-b", f2
			],
			capture_output = True,
			text = True # Get data as str instead of Bytes
		)
		if result.returncode != 0:
			same_files = False
			failed( f"Error while substracting {f1} by {f2}: ${result.returncode}" )
		if result.stderr:
			same_files = False
			failed( f"Got stderr while substracting {f1} by {f2}: " )
			print( result.stderr )
		if result.stdout:
			same_files = False
			failed( f"Missing data from {f2} compared to {f1}: " )
			print( result.stdout )
	return same_files

def cmp_eugene_logs( logfile_1: str, logfile_2: str ) -> bool:
	# Functions to define ignored lines (keep changing) and get only usable lignes
	# Ignore logs line of type INFO or DEBUG ( WARNING and ERROR are compared)
	def ignored_line( line: str ):
		return line.startswith( "[INFO]" ) or line.startswith( "[DEBUG]" )
	# Function to get the next non-ignored line in file
	def get_next_line( f: typing.TextIO ) -> str:
		line = f.readline()
		while ignored_line(line):
			line = f.readline()
		return line
	with open( logfile_1, 'r' ) as f1, open( logfile_2, 'r' ) as f2:
		# Important fact, readline return "\n" for an empty line, and "" for EOF, it's used to detect EOF here
		line_1: str = get_next_line(f1)
		line_2: str = get_next_line(f2)
		# Check all lines are same between files
		pattern_version = re.compile( "^EuGene rel. .* - Genus - mode Eukaryote -$" ) # To remove the versionning from comparison
		pattern_parfile_loaded = re.compile( "Parameters file .* loaded.$" ) # To remove paths from comparison
		pattern_dirpath = re.compile( "/.*/" )
		# Read files until one is EOF ( in python3 a string returns false only if it's empty and "\n" is not empty)
		while line_1 and line_2:
			# Ensure variable stuffs (paths and versions) don't impact the comparison
			line_1 = pattern_version.sub( "EuGene rel. #VERSION# - Genus - mode Eukaryote -", line_1 )
			line_2 = pattern_version.sub( "EuGene rel. #VERSION# - Genus - mode Eukaryote -", line_2 )
			line_1 = pattern_parfile_loaded.sub( "Parameters file #PAR_FILE# loaded.", line_1 )
			line_2 = pattern_parfile_loaded.sub( "Parameters file #PAR_FILE# loaded.", line_2 )
			line_1 = pattern_dirpath.sub( "#DIRPATH#/", line_1 )
			line_2 = pattern_dirpath.sub( "#DIRPATH#/", line_2 )
			# compare !
			if line_1 != line_2:
				return False
			line_1: str = get_next_line(f1)
			line_2: str = get_next_line(f2)
		# Check both files are at the end ( string return false only if empty so EOF here )
		if line_1 or line_2:
			return False
	return True

def running_test( name: str, direction: str, data_dir: str, run_dir: str ) -> bool:
	info( f"Running {name} {direction}" )
	output_prefix = f"{run_dir}/ANNOT.{direction}"

	run_data_dir = prepare_run_data( direction, data_dir, run_dir )

	par_file = f"{run_dir}/{direction}.par"
	
	with open( par_file, 'w' ) as outfile:
		subprocess.run(
			[
				"sed", f"s/EuGene.version .*/EuGene.version {egn_version}/",
				f"{run_data_dir}/annotationV1{direction}.par"
			],
			stdout = outfile,
			check=True
		)

	egn_cmd = [
		str(egn_bin),
		"-A", par_file,
		"-m", f"{run_data_dir}/eugene.mat",
		"-pg",
		"-O", output_prefix,
		f"{run_data_dir}/{name}" # Output prefix
	]
	egn_env = { "EUGENEDIR": f"{egn_dir}" }
	elapsed_time = -time.time()
	with open( f"{output_prefix}.stdout", 'w' ) as stdout_file, open( f"{output_prefix}.stderr", 'w' ) as stderr_file:
		egn_run = subprocess.run( egn_cmd, env=egn_env, stdout=stdout_file, stderr=stderr_file, cwd=run_dir )
		if egn_run.returncode != 0:
			failed( f"Execution failed with {egn_run.returncode} for {name} {direction}" )
			return False
	elapsed_time += time.time()

	no_fail = True

	if not cmp_gff3s( f"{run_data_dir}/ANNOT.{direction}{name}.gff3", f"{output_prefix}{name}.gff3" ):
		failed( f"Differents results on gff3 for {name} {direction}" )
		no_fail = False
	
	if not filecmp.cmp(	f"{run_data_dir}/ANNOT.{direction}{name}.eugene.stdout", f"{output_prefix}.stdout", shallow=False ):
		failed( f"Differents results on stdout for {name} {direction}" )
		no_fail = False

	if not cmp_eugene_logs( f"{run_data_dir}/ANNOT.{direction}{name}.eugene.stderr", f"{output_prefix}.stderr" ):
		failed( f"Differents results on stderr for {name} {direction}" )
		no_fail = False

	if no_fail:
		ok( f"Same results for {name} {direction}, executed in {int(elapsed_time)} seconds" )

	return no_fail

# Run core
if __name__ == "__main__":
	all_tests_ok: bool = False
	tests: dict = { } # Dict[ str, AsyncResult ]
	with mp.Pool( processes = NB_EUGENE_PARALLEL_RUNS ) as pool:
		with open( integrations_tests_list_file, newline='' ) as csvfile:
			reader = csv.DictReader( csvfile, delimiter=',' )
			for row in reader:
				test_name = row['TEST_NAME']
				data_dist = f"{test_dataset_srv}:{test_dataset_dir}/{test_name}"
				data_dir = (  script_dir / f"{test_name}/data" ).resolve()
				run_dir = ( script_dir / f"{test_name}/run-{isodate}" ).resolve()
				

				info( f"Preparing test '{test_name}'" )

				os.makedirs( data_dir, exist_ok = True )
				os.makedirs( run_dir, exist_ok = False ) # Avoid collisions on tests (like: you call the script twice without realising it)

				subprocess.run( [
					"rsync", "-r",
					"--ignore-existing",
					"--delete",
					"--compress",
					#"--progress",
					#"--info=progress2", # output total progress
					str(data_dist)+"/", # '+ "/"' to ensure to copy the directory content without the directory itself
					str(data_dir)
				], check = True )

				for direction in [ "FW", "RV" ]:
					tests[ f"{test_name} {direction}" ] = pool.apply_async(
						running_test,
						args = [
							test_name, direction,
							str(data_dir),
							str(run_dir)
						]
					)

		# Gather results of tests
		pool.close()
		pool.join()
		all_tests_ok = all( [ ( r.successful() and r.get() ) for r in tests.values()])


	if all_tests_ok:
		ok( "Passed ALL integration tests" )
		sys.exit( 0 )
	else:
		failed( "Did NOT passed all integration tests" )
		sys.exit( 1 )
