console.info( "Loading dag_representation.js: start" );

const main = document.getElementsByTagName("main")[0];
const egn_dag_tsv_input = document.getElementById( "load_egn_dag_tsv_input" );
egn_dag_tsv_input.accept = ".dag.tsv";
const window_start_pos_input = document.getElementById( "window_start_pos_input" );
const window_size_input = document.getElementById( "window_size_input" );
const tooltips = document.getElementById( "tooltips" );

function update_with_ui_values(){
	const dag_tsv = egn_dag_tsv_input.files.length ? egn_dag_tsv_input.files[0] : null;
	if( !( dag_tsv instanceof File ) ){
		main.innerHTML = "<p>No loaded file</p>";
		return;
	}
	update( dag_tsv, Number(window_start_pos_input.value), Number(window_size_input.value) );
}

async function update( dag_tsv, window_start_pos, window_size ){
	const data = await load_dag_data( dag_tsv );
	display_data( data, window_start_pos, window_start_pos+window_size );
	return;
}

const TSV_COLUMNS = [ "track", "backpoint", "next", "prev", "origin", "start_position", "cost" ];
const COLUMN_IS_NUMBER = [ true, false, false, false, false, true, true ];

const TRACKS = {
	0: "InitF1", 1: "InitF2", 2: "InitF3", 3: "InitR1", 4: "InitR2", 5: "InitR3",
	6: "SnglF1", 7: "SnglF2", 8: "SnglF3", 9: "SnglR1", 10: "SnglR2", 11: "SnglR3",
	12: "IntrF1", 13: "IntrF2", 14: "IntrF3", 15: "IntrR1", 16: "IntrR2", 17: "IntrR3",
	18: "TermF1", 19: "TermF2", 20: "TermF3", 21: "TermR1", 22: "TermR2", 23: "TermR3",
	24: "IntronF1", 25: "IntronF2", 26: "IntronF3", 27: "IntronR1", 28: "IntronR2", 29: "IntronR3",
	30: "IntronF2T", 31: "IntronF2A",
	32: "IntronF3TG", 33: "IntronF3TA", 34: "IntronF3TC", 35: "IntronF3AG",
	36: "IntronR3G", 37: "IntronR3A",
	38: "IntronR2GA", 39: "IntronR2AG", 40: "IntronR2AA", 41: "IntronR2CA", 42: "IntronR2GG",
	43: "InterGen",
	44: "UTR5F", 45: "UTR3F", 46: "UTR5R", 47: "UTR3R",
	48: "IntronU5F", 49: "IntronU5R", 50: "IntronU3F", 51: "IntronU3R",
	52: "RnaF", 53: "RnaR",
	// TODO
}

function parse_dag_data_line( line, is_first_line ){
	const rows = line.split('\t');

	// Check rows if first line
	if( is_first_line ){
		if( rows.length != TSV_COLUMNS.length || rows.some( function(element, index){ return element !== TSV_COLUMNS[index]; } ) ){
			console.error( "File has a bad header: ", rows, " instead of ", TSV_COLUMNS );
			return null;
		}
		return null;
	}

	// parse line
	const data = {};
	for( const [idx,field] of TSV_COLUMNS.entries() ){
		if( rows[idx] == "TRACK" ){
			data[field] = null;
			continue;
		}
		data[field] = COLUMN_IS_NUMBER[idx] ? Number(rows[idx]) : rows[idx];
	}

	return data;
}

async function load_dag_data( file, start_pos, end_pos ){
	main.innerHTML = "<p>Parsing file...</p>";

	if( !(file instanceof File) ){
		console.error( "Provided file is not a File: " + file );
		return;
	}

	const data = [];

	// Create progress indicators
	const progress_bar = document.createElement( "progress" );
	const remaining_msg = document.createElement( "p" );
	console.debug( "file size: " + file.size );
	progress_bar.max = file.size;
	progress_bar.value = 0;
	main.appendChild( progress_bar );
	main.appendChild( remaining_msg );
	const start_ms = performance.now();

	// Parsing file
	console.log( "Start parsing" );	
	const decoder = new TextDecoder( "utf-8" );
	const reader = file.stream().getReader();
	let current_line = "";
	let is_first_line = true;
	while(true){
		let { value: chunk, done: reader_done } = await reader.read();
		if( reader_done ){ break; }

		// Update progress interface
		progress_bar.value += chunk.length;
		const elapsed = performance.now() - start_ms;
		if( progress_bar.value > 0 ){
			const duration_total_ms_estimation = Math.floor( elapsed * progress_bar.max / progress_bar.value );
			const remaining_total_ms_estimation = duration_total_ms_estimation - elapsed;
			remaining_msg.innerHTML = `estimated remaining time: ${Math.floor(remaining_total_ms_estimation/1000)} s`;
		}

		// Parse this chunk
		let chunk_blocks = decoder.decode( chunk ).split("\n"); // Chunk data splitted by '\n'
		current_line += chunk_blocks.shift(); // The first block is the current line
		// Each additional block indicated a new line
		for( const block of chunk_blocks ){
			const entry =parse_dag_data_line( current_line, is_first_line );
			is_first_line = false;
			current_line = block;
			if( entry ){
				data.push( entry );
			}
		}
	}
	// Ensure to read last line
	if( current_line.length ){
		const entry = parse_dag_data_line( current_line, is_first_line );
		current_line = "";
		if( entry ){
			data.push( entry );
		}
	}

	return data;
}

function get_cost_color( cost, min_cost, max_cost ){
	if( cost == Number.NEGATIVE_INFINITY ){ return "hsl( 0 0% 50% )"; }
	// 0 is yellow, green is negative, red is positive
	let hue = 0;
	if( cost == 0.0 ){ hue = 60; }
	else if( cost < 0.0 ){ hue = 60 + 60*cost/min_cost; }
	else { hue = 60 - 60*cost/max_cost; }
	return( `hsl( ${hue} 100% 50% )` );
}

function display_data( data, start_pos, end_pos ){
	const tracks_row = [];
	for( const entry of data ){
		const track_id = entry["track"];
		if( !tracks_row.includes(track_id) ){
			tracks_row.push(track_id);
		}
	}

	const window_data = [];
	for( const entry of data ){
		const pos = entry["start_position"];
		if( pos >= start_pos && pos < end_pos ){
			window_data.push( entry );
		}
	}
	if( window_data.length == 0 ){
		main.innerHTML = "<p>Window has no data</p>";
		return;
	}
	const window_metadata = {
		nb_points: window_data.length,
		min_cost: window_data.map( (entry) => entry.cost ).reduce( (a,b) => Math.min(a,b) ),
		max_cost: window_data.map( (entry) => entry.cost ).reduce( (a,b) => Math.max(a,b) )
	}
	console.log( window_metadata );

	const ROW_HEIGHT = 10;
	const HEAD_WIDTH = 100;
	const ELEMENT_WIDTH = 1;
	const NB_ROWS = tracks_row.length;

	const style = getComputedStyle( document.documentElement );
	const COLORS = {
		background: [ style.getPropertyValue( "--color-bg-main" ), style.getPropertyValue("--color-bg-second" ) ],
		text: [ style.getPropertyValue( "--color-txt-main" ), style.getPropertyValue( "--color-txt-second" ) ],
		nucs: { 'a': "Green", 'c': "Blue", 'g': "Yellow", 't': "Red" }
	};

	main.innerHTML = "";
	const canvas = document.createElement( "canvas" );
	canvas.width = HEAD_WIDTH + (end_pos-start_pos)*ELEMENT_WIDTH;
	canvas.height = NB_ROWS * ROW_HEIGHT;
	const ctxt = canvas.getContext("2d");
	ctxt.clearRect( 0, 0, canvas.width, canvas.height );

	// Draw rows
	ctxt.font = `${ROW_HEIGHT}px monospace`;
	ctxt.fillStyle = COLORS.text[0];
	ctxt.textBaseline = "middle";
	ctxt.strokeStyle = COLORS.background[1];
	for( const [row_id,track_id] of tracks_row.entries() ){
		const row_center = (row_id+0.5)*ROW_HEIGHT;
		ctxt.fillText( TRACKS[track_id], 0, row_center, HEAD_WIDTH-ELEMENT_WIDTH ); // Restrict width with a column as a gap
		ctxt.beginPath();
		ctxt.moveTo( HEAD_WIDTH, row_center );
		ctxt.lineTo( canvas.width, row_center );
		ctxt.stroke();
	}

	// Draw contents
	for( const entry of window_data ){
		const pos_y = ( tracks_row[ entry["track"] ] + 0.5 ) * ROW_HEIGHT;
		const pos_x = HEAD_WIDTH + (entry["start_position"]	- start_pos)*ELEMENT_WIDTH + ELEMENT_WIDTH/2;

		// Draw origin link
		if( entry.origin ){
			const origin = data.find( (e) => e.backpoint === entry.origin );
			if( origin != undefined ){
				const origin_x = origin.start_position >= start_pos ? HEAD_WIDTH + (origin.start_position - start_pos)* ELEMENT_WIDTH + ELEMENT_WIDTH/2 : HEAD_WIDTH;
				const origin_y = ( tracks_row[origin.track] +0.5 ) * ROW_HEIGHT;

				const transition_x = pos_x - ELEMENT_WIDTH;
				ctxt.strokeStyle = COLORS.text[0];
				ctxt.beginPath();
				ctxt.moveTo( origin_x, origin_y );
				ctxt.lineTo( transition_x, origin_y );
				ctxt.lineTo( pos_x, pos_y );
				ctxt.stroke();
			}
		}

		// Draw current backpoint
		ctxt.strokeStyle = COLORS.text[0];
		ctxt.fillStyle = get_cost_color( entry.cost, window_metadata.min_cost, window_metadata.max_cost );
		ctxt.beginPath();
		ctxt.arc( pos_x, pos_y, ROW_HEIGHT/2, 0, 2*Math.PI )
		ctxt.fill();
		ctxt.stroke();
	}

	// Finished, display canvas
	main.appendChild( canvas );
	return;
}

egn_dag_tsv_input.addEventListener( "change", update_with_ui_values );
window_start_pos_input.addEventListener( "change", function(event){ update_with_ui_values(); } );
window_size_input.addEventListener( "change", function(event){
	window_start_pos_input.step = event.target.value;
	update_with_ui_values();
} );

console.info( "Loading dag_representation.js: done" );
update_with_ui_values();