console.info( "Loading debug_plotting.js: start" );

const main = document.getElementsByTagName("main")[0];
const egn_debug_file_input = document.getElementById( "load_egn_debug_file_input" );
const window_start_pos_input = document.getElementById( "window_start_pos_input" );
const window_size_input = document.getElementById( "window_size_input" );
const tooltips = document.getElementById( "tooltips" );

const EDGES = { forward: 22, reverse: 37, forward_no: 52, reverse_no: 67 }; // Edges and their first ID (signal) in the file
const SIGNALS = [
	"transcription_start", "transcription_stop", "start", "stop", "acceptor", "donor",
	"insertion_1", "insertion_2", "insertion_3",
	"deletion_1", "deletion_2", "deletion_3",
	"transcription_start_npc", "transcription_stop_npc"
];

let metadata = {};
let data = {};

function reset_window_data(){
	data = {
		pos: [],
		nt: [],
		contents: {
			metadata: { min_cost: -1, max_cost: 1 },
			exon_forward_1: [],
			exon_forward_2: [],
			exon_forward_3: [],
			exon_reverse_1: [],
			exon_reverse_2: [],
			exon_reverse_3: [],
			intron_f: [],
			intron_r: [],
			inter_g: [],
			utr5_f: [],
			utr5_r: [],
			utr3_f: [],
			utr3_r: [],
			intron_utr_f: [],
			intron_utr_r: [],
			rna_f: [],
			rna_r: [],
			uirf: [],
			uirr: []
		},
		edges_signals: {
			metadata: { min_cost: -1, max_cost: 1 }
		}
	}
	for( const edge of Object.keys(EDGES) ){
		data.edges_signals[edge] = {};
		for( const signal of SIGNALS ){
			data.edges_signals[edge][signal] = [];
		}
	}
}

function clear_data(){
	reset_window_data();
	metadata = {
		min_cost: -1, max_cost: 1,
		min_pos: Number.MAX_VALUE, max_pos: 1,
		file: undefined, file_index: {},
	};
}
clear_data(); // Call it, ensuring data and metadata are generated correctly at the first page load

function add_entry( entry ){
	data.pos.push( entry.pos );
	data.nt.push( entry.nt );
	for( const k of Object.keys( entry.contents ) ){
		const cost = entry.contents[k];
		data.contents[k].push( cost );
		if( cost != Number.NEGATIVE_INFINITY ){
			if( cost < data.contents.metadata.min_cost ){ data.contents.metadata.min_cost = cost; }
			if( cost > data.contents.metadata.max_cost ){ data.contents.metadata.max_cost = cost; }
		}
	} 

	for( const edge of Object.keys( entry.edges_signals ) ){
		for( const signal of Object.keys( entry.edges_signals[edge] ) ){
			const cost = entry.edges_signals[edge][signal];
			data.edges_signals[edge][signal].push( cost );
			if( cost != Number.NEGATIVE_INFINITY ){
				if( cost < data.edges_signals.metadata.min_cost ){ data.edges_signals.metadata.min_cost = cost; }
				if( cost > data.edges_signals.metadata.max_cost ){ data.edges_signals.metadata.max_cost = cost; }
			}
		}
	}
}

function update_metadata( entry, byte_start ){
	// Update min/max cost
	for( const v of Object.values( entry.contents ) ){
		if( !isFinite(v) ){ continue; }
		if( v < metadata.min_cost ){ metadata.min_cost = v; }
		if( v > metadata.max_cost ){ metadata.max_cost = v; }
	}
	for( const edge of Object.keys( entry.edges_signals ) ){
		for( const v of Object.values( entry.edges_signals[edge] ) ){
			if( !isFinite(v) ){ continue; }
			if( v < metadata.min_cost ){ metadata.min_cost = v; }
			if( v > metadata.max_cost ){ metadata.max_cost = v; }
		}
	}
	// Update index
	metadata.file_index[entry.pos] = byte_start;
	// Update positions range
	if( entry.pos < metadata.min_pos ){ metadata.min_pos = entry.pos; }
	if( entry.pos > metadata.max_pos ){ metadata.max_pos = entry.pos; }
}

function update_with_ui_values(){
	update( window_start_pos_input.value, window_size_input.value );
}

function update( window_start_pos, window_size ){
	if(  !( "file" in metadata ) || !( metadata.file instanceof File) ){
		main.innerHTML = "<p>No loaded file</p>";
		return;
	}	
	window_start_pos = Number(window_start_pos);
	window_size = Number(window_size);
	load_window_data( window_start_pos, window_start_pos+window_size );
}

const expected_header = [
	"pos", "nt", "EF1", "EF2", "EF3", "ER1", "ER2", "ER3", "IF", "IR", "IG", "U5F", "U5R", "U3F", "U3R", "IUF", "IUR", "RF", "RR", "UIRF", "UIRR",
	"FWD", "tSta", "tSto", "Sta", "Sto", "Acc", "Don", "Ins1", "Ins2", "Ins3", "Del1", "Del2", "Del3", "tStaNcp", "tStopNcp",
	"REV", "tSta", "tSto", "Sta", "Sto", "Acc", "Don", "Ins1", "Ins2", "Ins3", "Del1", "Del2", "Del3", "tStaNcp", "tStopNcp",
	"noF", "tSta", "tSto", "Sta", "Sto", "Acc", "Don", "Ins1", "Ins2", "Ins3", "Del1", "Del2", "Del3", "tStaNcp", "tStopNcp",
	"noR", "tSta", "tSto", "Sta", "Sto", "Acc", "Don", "Ins1", "Ins2", "Ins3", "Del1", "Del2", "Del3", "tStaNcp", "tStopNcp"
];

function read_cost( value, index_mode = true ){
	if( value == "-inf" ){ return Number.NEGATIVE_INFINITY; }
	else if( isNaN(value) ){
		console.error( "Encounter weird cost: ", value );
		return Number.NEGATIVE_INFINITY;
	}
	const v = Number(value);
	return Number(value);
}


function parse_egn_debug_line( line ){
	line = line.trim(); // Remove spaces at start and end of line (used by egn to align positions in text...)
	const elements = line.split( /\s+/ ); // The file is columnar, where the separator are consecutive spaces (because WHY FUCKING NOT ?)

	// It's an header line (which is typically set multiple line: first and last lines ) ?
	if( elements[0] == "pos" ){
		if( elements.toString() != expected_header.toString() ){
			throw new Error(
				"Header does not match the expected one: got " +
				String(elements) + " but expected " + String(expected_header)
			);
		}
		return null;
	}

	// Store this entry
	const line_data = {};
	line_data.pos = Number(elements[0]);
	line_data.nt = elements[1];

	// Add contents
	line_data.contents = {};
	line_data.contents.exon_forward_1 = read_cost(elements[2]);
	line_data.contents.exon_forward_2 = read_cost(elements[3]);
	line_data.contents.exon_forward_3 = read_cost(elements[4]);
	line_data.contents.exon_reverse_1 = read_cost(elements[5]);
	line_data.contents.exon_reverse_2 = read_cost(elements[6]);
	line_data.contents.exon_reverse_3 = read_cost(elements[7]);
	line_data.contents.intron_f = read_cost(elements[8]);
	line_data.contents.intron_r = read_cost(elements[9]);
	line_data.contents.inter_g = read_cost(elements[10]);
	line_data.contents.utr5_f = read_cost(elements[11]);
	line_data.contents.utr5_r = read_cost(elements[12]);
	line_data.contents.utr3_f = read_cost(elements[13]);
	line_data.contents.utr3_r = read_cost(elements[14]);
	line_data.contents.intron_utr_f = read_cost(elements[15]);
	line_data.contents.intron_utr_r = read_cost(elements[16]);
	line_data.contents.rna_f = read_cost(elements[17]);
	line_data.contents.rna_r = read_cost(elements[18]);
	line_data.contents.uirf = read_cost(elements[19]);
	line_data.contents.uirr = read_cost(elements[20]);

	// Add signals
	line_data.edges_signals = {};
	for( const [edge,first_id] of Object.entries(EDGES) ){
		let id = first_id;
		line_data.edges_signals[edge] = {};
		for( const signal of SIGNALS ){
			line_data.edges_signals[edge][signal] = read_cost(elements[id]);
			id++;
		}
	}

	return line_data;
};

async function load_egn_debug_file ( file ){
	main.innerHTML = "<p>Clearing current data...</p>";
	clear_data();
	window_start_pos_input.min = 1;
	window_start_pos_input.max = 1;
	window_start_pos_input.value = 1;
	window_start_pos_input.step = window_size_input.value;

	main.innerHTML = "<p>Parsing file...</p>";
	metadata.file = file;
	const progress_bar = document.createElement( "progress" );
	const remaining_msg = document.createElement( "p" );
	progress_bar.max = file.size;
	progress_bar.value=0;
	main.appendChild( progress_bar );
	main.appendChild( remaining_msg );

	let bytes_parsed = 0;
	console.log( "Start parsing" );
	const start_ms = performance.now();
	const decoder = new TextDecoder("utf-8");
	const reader = file.stream().getReader();
	let current_line = "";
	let current_line_start_byte = 0;
	while(true){
		let { value: chunk, done: reader_done } = await reader.read();
		if( reader_done ){ break; }
		// Update waiting interface
		progress_bar.value += chunk.length;
		const elapsed = performance.now() - start_ms;
		if( progress_bar.value > 0 ){
			const duration_total_ms_estimation = Math.floor( elapsed * progress_bar.max / progress_bar.value );
			const remaining_total_ms_estimation = duration_total_ms_estimation - elapsed;
			remaining_msg.innerHTML = `estimated remaining time: ${get_human_timing(remaining_total_ms_estimation)}`;
		}
		// Parse this chunk
		let chunk_blocks = decoder.decode( chunk ).split("\n"); // chunk data splitted by "\n"
		// The first block is the CURRENT line
		current_line += chunk_blocks.shift(); // Use shift to remove it from the list at the same time
		// Each additional block indicates a new line
		for( const block of chunk_blocks ){
			const entry = parse_egn_debug_line( current_line );
			if( entry != null ){
				update_metadata( entry, current_line_start_byte );
			}
			current_line_start_byte += current_line.length + 1; // +1 for the '\n'
			current_line = block;
		}
	}
	// Ensure last line is read
	if( current_line.length ){
		const entry = parse_egn_debug_line( current_line );
		if( entry != null ){
			update_metadata( entry, current_line_start_byte );
		}
		current_line_start_byte += current_line.length + 1;
		current_line = "";
	}
	window_start_pos_input.min = metadata.min_pos;
	window_start_pos_input.max = metadata.max_pos - Number(window_size_input.value);
	window_start_pos_input.value = metadata.min_pos;
	console.log( "Finished parsing file, in", get_human_timing( performance.now() - start_ms )  );
	update_with_ui_values();
	return;
};

async function load_window_data( pos_start, pos_end ){
	reset_window_data();

	if( pos_start < metadata.min_pos ){
		pos_start = metadata.min_pos;
		pos_end = pos_start + Number(window_size_input.value);
		window_start_pos_input.value = pos_start;
	}
	if( pos_start > metadata.max_pos ){
		pos_start = Math.max( metadata.max_pos - Number(window_size_input.value)+1, metadata.min_pos );
		pos_end = pos_start + Number(window_size_input.value);
		window_start_pos_input.value = pos_start;
	}

	console.log( "Start loading" );
	const bytes_start = metadata.file_index[pos_start];
	const bytes_end = pos_end in metadata.file_index ? metadata.file_index[pos_end] : metadata.file.size;
	main.innerHTML = "<p>Loading window...</p>";
	const progress_bar = document.createElement( "progress" );
	const remaining_msg = document.createElement( "p" );
	progress_bar.max = bytes_end - bytes_start;
	progress_bar.value=0;
	main.appendChild( progress_bar );
	main.appendChild( remaining_msg );

	const start_ms = performance.now();
	const decoder = new TextDecoder("utf-8");
	const reader = metadata.file.slice( bytes_start, bytes_end ).stream().getReader();
	let current_line = "";
	while(true){
		let { value: chunk, done: reader_done } = await reader.read();
		if( reader_done ){ break; }

		// Update waiting interface
		progress_bar.value += chunk.length;
		const elapsed = performance.now() - start_ms;
		if( progress_bar.value > 0 ){
			const duration_total_ms_estimation = Math.floor( elapsed * progress_bar.max / progress_bar.value );
			const remaining_total_ms_estimation = duration_total_ms_estimation - elapsed;
			remaining_msg.innerHTML = `estimated remaining time: ${get_human_timing(remaining_total_ms_estimation)}`;
		}

		// Parse this chunk
		let chunk_blocks = decoder.decode( chunk ).split("\n"); // chunk data splitted by "\n"
		current_line += chunk_blocks.shift(); // Use shift to remove it from the list at the same time
		// Each additional block indicates a new line
		for( const block of chunk_blocks ){
			const entry = parse_egn_debug_line( current_line );
			if( entry != null ){
				add_entry( entry );
			}
			current_line = block;
		}
	}
	// Ensure last line is read
	if( current_line.length ){
		const entry = parse_egn_debug_line( current_line );
		if( entry != null ){
			add_entry( entry );
		}
		current_line = "";
	}
	console.log( "Finished loading window data, in", get_human_timing( performance.now() - start_ms )  );
	display_data( pos_start, pos_end - pos_start );
}

function get_human_timing( time_ms ){
	let time = Math.floor( time_ms / 1000 ); // seconds
	if( time < 60 ){ return `${time} s`; }
	time = Math.floor(time/60);
	if( time < 60 ){ return `${time} m`; }
	time = Math.floor(time/60);
	if( time < 24 ){ return `${time} h`; }
	time = Math.floor(time/24);
	return `${time} d`;
}

async function display_data( start_pos = 1, window_size = 100 ){
	// Ensure par pameters a numbers (numbers from inputs are string...)
	start_pos = Number(start_pos);
	window_size = Number(window_size);

	const nb_nucs = data.pos.length;
	// Clear main
	main.innerHTML = "";
	if( nb_nucs == 0 ){
		main.innerHTML = "<p>no data loaded</p>";
		return;
	}

	const style = getComputedStyle( document.documentElement );
	const COLORS = {
		background: [ style.getPropertyValue( "--color-bg-main" ), style.getPropertyValue("--color-bg-second" ) ],
		text: [ style.getPropertyValue( "--color-txt-main" ), style.getPropertyValue( "--color-txt-second" ) ],
		nucs: { 'a': "Green", 'c': "Blue", 'g': "Yellow", 't': "Red" }
	};

	const ELEMENT_WIDTH = 10;
	const HEAD_WIDTH = 100;
	const ROW_HEIGHT = 10;

	// Add colors scales and position representation
	main.appendChild( get_costs_scale_canvas( 800, 1.2*ROW_HEIGHT, HEAD_WIDTH, ELEMENT_WIDTH, COLORS ) );
	main.appendChild( get_window_position_canvas( 800, ROW_HEIGHT, HEAD_WIDTH, ELEMENT_WIDTH, COLORS, start_pos, window_size ) );

	const nb_id = data.pos.length;
	const CONTENTS_ROWS = [
		"Exon Forward 1", "Exon Forward 2", "Exon Forward 3",
		"Exon Reverse 1", "Exon Reverse 2", "Exon Reverse 3",
		"Intron F", "Intron R", "InterGenic", "UTR5 F", "UTR5 R", "UTR3 F", "UTR3 R",
		"Intron UTR F", "Intron UTR R", "RNA F", "RNA R", "UIRF", "UIRR"
	];
	const CONTENTS_ROWS_STORAGE = [
		"exon_forward_1", "exon_forward_2", "exon_forward_3",
		"exon_reverse_1", "exon_reverse_2", "exon_reverse_3",
		"intron_f", "intron_r", "inter_g", "utr5_f", "utr5_r", "utr3_f", "utr3_r",
		"intron_utr_f", "intron_utr_r", "rna_f", "rna_r", "uirf", "uirr"
	];
	const EDGES = [ "forward", "reverse", "forward_no", "reverse_no"];
	const SIGNAL_ROWS = [
		"Transcription Start", "Transcription Stop",
		"Start", "Stop", "Acceptor", "Donor",
		"Insertion 1", "Insertion 2", "Insertion 3",
		"Deletion 1", "Deletion 2", "Deletion 3",
		"Transcription Start NPC", "Transcription Stop NPC"
	];
	const SIGNALS_ROWS_STORAGE = [
		"transcription_start", "transcription_stop",
		"start", "stop", "acceptor", "donor",
		"insertion_1", "insertion_2", "insertion_3",
		"deletion_1", "deletion_2", "deletion_3",
		"transcription_start_npc", "transcription_stop_npc"
	]
	const NB_ROWS = 2 + (CONTENTS_ROWS.length+1) + EDGES.length*(SIGNAL_ROWS.length+1);
	const canvas = document.createElement( "canvas" )
	canvas.width = HEAD_WIDTH + nb_id*ELEMENT_WIDTH;
	canvas.height = NB_ROWS*ROW_HEIGHT;

	const ctxt = canvas.getContext("2d");
	ctxt.clearRect( 0, 0, canvas.width, canvas.height );

	// Draw rows names
	ctxt.font = `${ROW_HEIGHT}px monospace`;
	ctxt.fillStyle = COLORS.text[0];
	ctxt.textBaseline = "middle";
	let row_id = 0.5;
	ctxt.fillText( "Position", 0, (row_id++)*ROW_HEIGHT, HEAD_WIDTH-ELEMENT_WIDTH); // Restrict width without a column as gap
	ctxt.fillText( "Nucleotide", 0, (row_id++)*ROW_HEIGHT, HEAD_WIDTH-ELEMENT_WIDTH); // Restrict width without a column as gap
	ctxt.textAlign = "center";
	ctxt.fillText( "CONTENTS", canvas.width/2, (row_id++)*ROW_HEIGHT, canvas.width/2); // Restrict width without a column as gap
	ctxt.textAlign = "left";
	for( const [id,row_name] of Object.entries( CONTENTS_ROWS ) ){
		ctxt.fillText( row_name, 0, (row_id++)*ROW_HEIGHT, HEAD_WIDTH-ELEMENT_WIDTH); // Restrict width without a column as gap
	}
	for( const edge of EDGES ){
		ctxt.textAlign = "center";
		ctxt.fillText( `${edge.toUpperCase()} SIGNALS`, canvas.width/2, (row_id++)*ROW_HEIGHT, canvas.width/2); // Restrict width without a column as gap
		ctxt.textAlign = "left";
		for( const [id,row_name] of Object.entries( SIGNAL_ROWS ) ){
			ctxt.fillText( row_name, 0, (row_id++)*ROW_HEIGHT, HEAD_WIDTH-ELEMENT_WIDTH); // Restrict width without a column as gap
		}
	}

	// Draw positions
	ctxt.fillStyle = COLORS.text[0];
	ctxt.textAlign = "left";
	ctxt.fillText( data.pos[0], HEAD_WIDTH, 0.5*ROW_HEIGHT );
	ctxt.textAlign = "right";
	ctxt.fillText( data.pos[data.pos.length-1], HEAD_WIDTH+nb_id*ELEMENT_WIDTH, 0.5*ROW_HEIGHT );

	ctxt.strokeStyle = COLORS.background[0];
	function draw_cell( pos_x, pos_y, color ){
		ctxt.fillStyle = color;
		ctxt.fillRect( pos_x, pos_y, ELEMENT_WIDTH, ROW_HEIGHT );
		ctxt.strokeRect( pos_x, pos_y, ELEMENT_WIDTH, ROW_HEIGHT );
	}

	for( let i = 0; i < nb_id; i++ ){
		const pos_x = HEAD_WIDTH + i*ELEMENT_WIDTH;

		// Draw nucs
		let row_id = 1;
		draw_cell( pos_x, row_id*ROW_HEIGHT, COLORS.nucs[ data.nt[i] ] );
		ctxt.font = `${ROW_HEIGHT*0.8}px monospace`;
		ctxt.fillStyle = COLORS.text[1];
		ctxt.textAlign = "center";
		ctxt.fillText( data.nt[i], pos_x + 0.5*ELEMENT_WIDTH, (row_id+0.5)*ROW_HEIGHT, ELEMENT_WIDTH );
		row_id++;

		// Draw CONTENTS
		row_id ++; // +1 As we need to jump to title line
		for( const name of CONTENTS_ROWS_STORAGE ){
			const pos_y = row_id*ROW_HEIGHT;
			row_id++;
			const value = data.contents[name][i];
			draw_cell( pos_x, pos_y, get_cost_color( value, data.contents.metadata.min_cost, data.contents.metadata.max_cost ) );
		}

		// Draw SIGNALS
		for( const edge of EDGES ){
			row_id++; // Jump the section title line
			for( const name of SIGNALS_ROWS_STORAGE ){
				const pos_y = row_id*ROW_HEIGHT;
				row_id++;
				const value = data.edges_signals[edge][name][i];
				draw_cell( pos_x, pos_y, get_cost_color( value, data.edges_signals.metadata.min_cost, data.edges_signals.metadata.max_cost ) );
			}
		}
	}
	canvas.data = { row_height: ROW_HEIGHT, head_width: HEAD_WIDTH, element_width: ELEMENT_WIDTH, contents_rows: CONTENTS_ROWS_STORAGE, edges: EDGES, signals_rows: SIGNALS_ROWS_STORAGE };
	canvas.addEventListener( "mousemove", mousemove_on_display );
	main.appendChild( canvas );

	return;
}

function get_costs_scale_canvas( canvas_width, row_height, head_width, head_right_padding, COLORS ){
	const canvas = document.createElement( "canvas" );
	canvas.width = canvas_width;
	canvas.height = 2*row_height;
	const ctxt = canvas.getContext("2d");

	ctxt.fillStyle = COLORS.text[0];
	ctxt.font = `${row_height}px monospace`;
	ctxt.textAlign = "left";
	ctxt.textBaseline = "middle";
	ctxt.fillText( "contents cost", 0, 0.5*row_height, head_width - head_right_padding );
	ctxt.fillText( "signals cost", 0, 1.5*row_height, head_width - head_right_padding );

	function add_scale( row_id, min_cost, max_cost ){
		const gradient = ctxt.createLinearGradient( head_width, row_id*row_height, canvas.width, (row_id+1)*row_height ); 
		const pos_0 = -min_cost/(max_cost-min_cost);
		gradient.addColorStop( 0, get_cost_color(min_cost, min_cost, max_cost) );
		gradient.addColorStop( pos_0, get_cost_color( 0, min_cost, max_cost ) ); // Ensure gradient respect middle (yellow)
		gradient.addColorStop( 1, get_cost_color(max_cost, min_cost, max_cost) );
		ctxt.fillStyle = gradient;
		ctxt.strokeStyle = COLORS.background[0];
		ctxt.fillRect( head_width, row_id*row_height, canvas.width-head_width, row_height );
		ctxt.strokeRect( head_width, row_id*row_height, canvas.width-head_width, row_height );
		ctxt.fillStyle = COLORS.text[1];
		ctxt.font = `${row_height*0.8}px monospace`;
		ctxt.textBaseline = "middle";
		ctxt.textAlign = "left";
		ctxt.fillText( min_cost, head_width+1, (row_id+0.5)*row_height );
		ctxt.textAlign = "middle";
		ctxt.fillText( 0, pos_0*(canvas.width-head_width)+head_width, (row_id+0.5)*row_height );
		ctxt.textAlign = "right";
		ctxt.fillText( max_cost, canvas.width-1, (row_id+0.5)*row_height );
	}
	add_scale( 0, data.contents.metadata.min_cost, data.contents.metadata.max_cost );
	add_scale( 1, data.edges_signals.metadata.min_cost, data.edges_signals.metadata.max_cost );

	return canvas;
}

function get_window_position_canvas( canvas_width, row_height, head_width, head_right_padding, COLORS, start_pos, window_size ){
	const canvas = document.createElement( "canvas" );
	canvas.width = canvas_width;
	canvas.height = row_height;
	const ctxt = canvas.getContext("2d");

	ctxt.fillStyle = COLORS.text[0];
	ctxt.font = `${row_height}px monospace`;
	ctxt.textAlign = "left";
	ctxt.textBaseline = "middle";
	ctxt.fillText( "window position", 0, 0.5*row_height, head_width - head_right_padding );

	const rect_width = canvas_width - head_width;

	ctxt.fillStyle = COLORS.background[1];
	ctxt.strokeStyle = COLORS.background[0];
	ctxt.fillRect( head_width, 0, rect_width, row_height );
	ctxt.strokeRect( head_width, 0, rect_width, row_height );

	const draw_factor = rect_width / (metadata.max_pos-metadata.min_pos)
	if( window_size*draw_factor > 1 ){
		ctxt.fillStyle = COLORS.text[0];
		ctxt.strokeStyle = COLORS.background[0];
		ctxt.fillRect( head_width + (start_pos-metadata.min_pos)*draw_factor, 0, window_size*draw_factor, row_height );
		ctxt.strokeRect( head_width + (start_pos-metadata.min_pos)*draw_factor, 0, window_size*draw_factor, row_height );
	} else {
		ctxt.fillStyle = COLORS.text[0];
		ctxt.fillRect( head_width + (start_pos-metadata.min_pos)*draw_factor, 0, 1, row_height );
	}

	return canvas;
}

function get_cost_color( cost, min_cost, max_cost ){
	if( cost == Number.NEGATIVE_INFINITY ){ return "hsl( 0 0% 50% )"; }
	// 0 is yellow, green is negative, red is positive
	let hue = 0;
	if( cost == 0.0 ){ hue = 60; }
	else if( cost < 0.0 ){ hue = 60 + 60*cost/min_cost; }
	else { hue = 60 - 60*cost/max_cost; }
	return( `hsl( ${hue} 100% 50% )` );
}

// function use to display data when mouse is move on the main display canvas
function mousemove_on_display( event ){
	if( !(event.target instanceof HTMLCanvasElement) ){
		console.error( "event target is not a canvas: ", event.target );
		tooltips_hide();
		return;
	}
	const canvas = event.target;
	if( !('data' in canvas) ){
		console.error( "Provided canvas is missing data field: ", canvas );
		tooltips_hide();
		return;
	}

	// Get the current data position and data row (contents, signal, ...)
	const padding = { left: 5, top: 5 }; // TODO: automatic finding: document.defaultView.getComputedStyle( main.firstChild ).getPropertyValue("padding-left")
	const canvas_rect = canvas.getBoundingClientRect();
	//const x = event.pageX - canvas_rect.left - padding.left;
	//const y = event.pageY - canvas_rect.top - padding.top;
	const x = event.offsetX - padding.left;
	const y = event.offsetY - padding.top;
	const row_height = canvas.data.row_height;
	const head_width = canvas.data.head_width;
	const element_width = canvas.data.element_width;
	if( x < canvas.data.head_width ){
		tooltips_hide();
		return;
	}
	const col_id = Math.floor( (x - head_width)/element_width );
	const row_id = Math.floor( y/row_height );

	if( col_id >= data.pos.length ){
		tooltips_hide();
		return;
	}

	let row_name = undefined;
	let value = undefined;

	const position = data.pos[col_id];
	const nucleotide = data.nt[col_id]

	if( row_id <= 2 ){
		tooltips_hide();
		return;
	} else if( row_id < 3+canvas.data.contents_rows.length ){ // contents
		row_name = canvas.data.contents_rows[ row_id-3 ];
		value = data.contents[row_name][col_id];
	} else {
		let i = row_id - 3 - canvas.data.contents_rows.length;
		const edge_id = Math.floor( i / (canvas.data.signals_rows.length+1) );
		const edge_name = canvas.data.edges[edge_id];
		i = i % (canvas.data.signals_rows.length+1);
		if( i == 0 ){ tooltips_hide(); return; }
		i -= 1; // Less on because of the header row
		const signal_name = canvas.data.signals_rows[i];
		row_name = `${edge_name} ${signal_name}`;
		value = data.edges_signals[edge_name][signal_name][col_id];
	}

	// Display tooltips
	tooltips.style.visibility = "visible";
	tooltips.style.left = `${event.pageX+20}px`;
	tooltips.style.top = `${event.pageY+20}px`;
	tooltips.innerHTML = "";
	const table = document.createElement( "table" );
	table.innerHTML = `<tr><th>position</th><td>${position}</td></tr>`;
	table.innerHTML += `<tr><th>nucleotide</th><td>${nucleotide}</td></tr>`;
	table.innerHTML += `<tr><th>name</th><td>${row_name}</td></tr>`;
	table.innerHTML += `<tr><th>value</th><td>${value}</td></tr>`;
	tooltips.appendChild( table );
}

function tooltips_hide() { tooltips.style.visibility = "hidden"; }

egn_debug_file_input.addEventListener( "change", function(event){ load_egn_debug_file( event.target.files[0] ) }, false );
egn_debug_file_input.accept = ".egn.debug";

window_start_pos_input.addEventListener( "change", function(event){ update_with_ui_values(); } );
window_size_input.addEventListener( "change", function(event){
	window_start_pos_input.step = event.target.value;
	window_start_pos_input.max = metadata.max_pos - Number(event.target.value);

	update_with_ui_values();
} );

console.info( "Loading debug_plotting.js: done" );
update_with_ui_values();
